from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import dominate
from dominate.tags import *
from nipype import logging as nlogging
from nilearn import plotting
from bs4 import BeautifulSoup

def guess_type(filepath):
    """
    Return the mimetype of a file, given it's path.

    This is a wrapper around two alternative methods - Unix 'file'-style
    magic which guesses the type based on file content (if available),
    and simple guessing based on the file extension (eg .jpg).

    :param filepath: Path to the file.
    :type filepath: str
    :return: Mimetype string.
    :rtype: str
    """
    try:
        import magic  # python-magic
        return magic.from_file(filepath, mime=True)
    except ImportError:
        import mimetypes
        return mimetypes.guess_type(filepath)[0]

def file_to_base64(filepath):
    """
    Returns the content of a file as a Base64 encoded string.

    :param filepath: Path to the file.
    :type filepath: str
    :return: The file content, Base64 encoded.
    :rtype: str
    """
    import base64
    if os.path.exists(filepath):
        with open(filepath, 'rb') as f:
            encoded_str = base64.b64encode(f.read())
        return encoded_str.decode('utf-8')
    else:
        return "error: image not located"


def make_html_images_inline(in_filepath, out_filepath):
    """
    Takes an HTML file and writes a new version with inline Base64 encoded
    images.

    :param in_filepath: Input file path (HTML)
    :type in_filepath: str
    :param out_filepath: Output file path (HTML)
    :type out_filepath: str
    """
    basepath = os.path.split(in_filepath.rstrip(os.path.sep))[0]
    soup = BeautifulSoup(open(in_filepath, 'r'), 'html.parser')

    # add 
    new_style = soup.new_tag('style')
    new_style.attrs['type']="text/css"

    for head in soup.find_all('head'):
        for link in head.find_all('link'):
            if link.has_attr("rel"):
                if link["rel"][0] == "stylesheet":
                    if link.has_attr("href"):
                        styleref = link["href"]
                        if os.path.exists(styleref):
                            with open(styleref, 'r') as fileid:
                                styletext=fileid.read()
                            styletext = styletext.replace('\t','').replace('\n','')
                            link.decompose()
                            new_style.string = styletext
                            head.append(new_style)


    for img in soup.find_all('img'):
        img_path = os.path.join(basepath, img.attrs['src'])
        mimetype = guess_type(img_path)
        img.attrs['src'] = \
            "data:%s;base64,%s" % (mimetype, file_to_base64(img_path))

    for obj in soup.find_all('object'):
        if obj.has_attr('type') and obj.has_attr('data'):
            if obj["type"] == "image/svg+xml":
                svgdata_file=obj["data"]
                if os.path.exists(svgdata_file):
                    with open(svgdata_file,'r') as infile:
                        svgdata=infile.read()
                    svgclean=svgdata.replace('\n','').replace('\t','').replace('"',"'")
                    #svgclean="<img src=\"image/svg+xml;utf8,"+svgclean + "\"></img>"
                    #new_img = soup.new_tag('img')
                    #new_img.attrs['src']="image/svg+xml;utf8,{}".format(BeautifulSoup(svgclean,'html.parser'))
                    obj.insert_after(BeautifulSoup(svgclean,'html.parser'))
                    #obj.insert_after(new_img)
                    obj.decompose()
            elif obj["type"] == "image/gif":
                svgdata_file=obj["data"]
                if os.path.exists(svgdata_file):
                    mimetype = guess_type(svgdata_file)
                    new_img = soup.new_tag('img')
                    new_img.attrs['src'] = "data:%s;base64,%s" % (mimetype, file_to_base64(svgdata_file))
                    obj.insert_after(new_img)
                    obj.decompose()
                        

    for svgobj in soup.find_all('svg'):
        svgobj.attrs["class"] = "svg-reportlet"

    allsoup=str(soup)

    with open(out_filepath, 'w') as of:
        of.write(str(allsoup))

def create_document(title, stylesheet=None, script=None):
    doc = dominate.document(title = title)
    if stylesheet is not None:
        with doc.head:
            link(rel='stylesheet',href=stylesheet)
    if script is not None:
        with doc.head:
            script(type='text/javascript',src=script)
    with doc:
        with div(id='header',cls="header"):
            h1(title)
        with div(id='timestamp',cls="header_timestamp"):
            p('Report generated on {}'.format(datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%dT%H:%M:%S.%f')) )
    return doc

def create_section(doc, divid, divclass, captiontext):
    with doc:
        if divclass is None:
            d = div(id=divid)
        else:
            d = div(id=divid, cls=divclass)
        with d:
            h2(captiontext)
    return doc


def create_table(doc, divid, divclass, tabid, tabclass, headers, captiontext, reportlist):
    with doc:
        if divclass is None:
            d = div(id=divid)
        else:
            d = div(id=divid, cls=divclass)
        with d:
            h3(captiontext)
            if tabclass is None:
                t = table(id = tabid)
            else:
                t = table(id = tabid, cls = tabclass)
            with t:
                with thead():
                    with tr():
                        for header in headers:
                            th(header)
                with tbody():
                    for listitem in reportlist:
                        with tr():
                            for itemvalue  in listitem:
                                td(itemvalue)
    return doc

def create_float_table(tabid, tabclass, headers, reportlist):
    if tabclass is None:
        t = table(id = tabid)
    else:
        t = table(id = tabid, cls = tabclass)
    with t:
        with thead():
            with tr():
                for header in headers:
                    th(header)
        with tbody():
            for listitem in reportlist:
                with tr():
                    for itemvalue  in listitem:
                        td(itemvalue)
    return t

def add_image(doc, divid, divclass, captiontext, image):
    with doc:
        if divclass is None:
            d = div(id=divid)
        else:
            d = div(id=divid, cls=divclass)
        with d:
            h3(captiontext)
            img(src=image)
    return doc

def add_float_image(imgid, imgclass, image):
    if imgclass is None:
        m = img(id=imgid, src=image)
    else:
        m = img(id=imgid, cls=imgclass, src=image)

    return m



def formatDateTime(row, fmt):
    dt = datetime.datetime.strptime(row['datetime'],'%Y-%m-%dT%H:%M:%S.%f')
    return datetime.datetime.strftime(dt, fmt)

def returnAverage(row):
    if len(str(row['value']).split(',')) > 1:
        row_values= [float(x) for x in row['value'].replace('[','').replace(']','').replace(' ','').split(',')]
        return np.mean(np.asarray(row_values))
    else:
        return row['value']


def writeROIimage(mask_rois, threeDfile, image ):
    combo = None
    for roi in mask_rois:
        roiimg = nib.load(roi)
        roidata = roiimg.get_fdata()
        if combo is None:
            combo = roidata
            combo_affine = roiimg.affine
            combo_header = roiimg.header
        else:
            combo=np.add(combo,roidata)

    funcimg = nib.load(threeDfile)
    if len(funcimg.header.get_data_shape()) > 3:
        funclist = nib.funcs.four_to_three(funcimg)
        threeDfile = funclist[0]

    combo_img = nib.Nifti1Image(combo, combo_affine, combo_header)

    display=plotting.plot_roi(combo_img, bg_img=threeDfile)
    display.savefig(image)


def writeStatImage(threeDfile, image, dmode='ortho'):
    display=plotting.plot_stat_map(threeDfile, display_mode=dmode)
    display.savefig(image)


def createRegisteredImage(image_list,image_png, only_first = False, bg_image_file = None, levellist=[0.5],colors=['r','g','b','y','m','c'],roi_index=None):

    contour_list=[]
    for image_file in image_list:
        contour_img = nib.load(image_file)
        if len(contour_img.header.get_data_shape()) > 3:
            expand_img = nib.funcs.four_to_three(contour_img)
            if only_first:
                contour_list.append(expand_img[0])
            else:
                contour_list.extend(expand_img)
        else:
            contour_list.append(contour_img)

    bg_img = None
    contours = []
    if bg_image_file:
        bg_img = nib.load(bg_image_file)
        contours = contour_list
        if roi_index:
            roi_index = min(roi_index,len(contour_list) - 1)
    else:
        if contour_list:
            bg_img = contour_list[0]
            if len(contour_list) > 1:
                contours = contour_list[1:]
                if roi_index:
                    roi_index = min(roi_index,len(contours) - 1)
            else:
                roi_index = None
        else:
            roi_index=None

    if bg_img:
        writeContourImages(bg_img,contours, image_png, levellist=levellist,colors=colors,roi_index=roi_index)
    else:
        print("Error - no image files provided")

            
def writeContourImages(bg_img,contours, png_file, levellist=[0.5],colors=['r','g','b','y','m','c'],roi_index=None):
        if roi_index is not None:
            mycanvas=plotting.plot_roi(contours[roi_index], bg_img=bg_img)
        else:
            mycanvas=plotting.plot_anat(bg_img)
        colorcount=0
        colorsize=len(colors)
        for contour_img in contours:
            colornum=colorcount%colorsize
            mycanvas.add_contours(contour_img,levels=levellist,colors=colors[colornum])
            colorcount=colorcount+1
      
        if not os.path.exists(os.path.dirname(png_file)):
            os.makedirs(os.path.dirname(png_file),exist_ok=True)
        mycanvas.savefig(png_file)
    

def AddDataSection(doc, doc_dict):

    headers = doc_dict["headers"]
    rows = doc_dict["rows"]
    title = doc_dict["title"]
    divid = doc_dict["divid"]

    if "tabid" in doc_dict.keys():
        tabid = doc_dict["tabid"]
    else:
        tabid = divid

    if "tabclass" in doc_dict.keys():
        tabclass = doc_dict["tabclass"]
    else:
        tabclass = None

    if "divclass" in doc_dict.keys():
        divclass = doc_dict["divclass"]
    else:
        divclass = None

    doc = create_table(doc, divid, divclass, tabid, tabclass, headers, title, rows)

    return doc

def AddRegSection(doc,doc_dict,only_first = False, bg_image_file = None, levellist=[0.5],colors=['r','g','b','y','m','c'],roi_index=None):

    image_list = doc_dict["image_list"]
    image_png = doc_dict["image_png"]
    title = doc_dict["title"]
    divid = doc_dict["divid"]

    if "divclass" in doc_dict.keys():
        divclass = doc_dict["divclass"]
    else:
        divclass = None

    createRegisteredImage(image_list,image_png,only_first = only_first, bg_image_file = bg_image_file, levellist=levellist,colors=colors,roi_index=roi_index)

    doc = add_image(doc, divid, divclass, title, image_png)

    return doc


def addRoiExtractProcReport(doc_dict):

    stylesheet = doc_dict["stylesheet"]
    title = doc_dict["title"]
    image_dir = doc_dict["image_dir"]
    output_dir = doc_dict["output_dir"] 
    mask = doc_dict["mask"] 
    atlas = doc_dict["atlas"] 
    measure = doc_dict["measure"]
    pipeline = doc_dict["pipeline"]
    metadata = doc_dict["metadata"]

    doc = create_document(title, stylesheet)
    with doc:
        with div(id='links',cls='contents').add(ul()):
            h2('Contents')
            li(a('Measure Registration',href='#measureregistration'))
            nested=ul()
            with nested:
                if mask:
                    for i in ['mask_measure', 'atlas_measure','mask_atlas_measure']:
                        li(a(i.title(), href='#%s' % i))
                else:
                    for i in ['atlas_measure']:
                        li(a(i.title(), href='#%s' % i))

            li(a('Measures Info',href='#measuresinfo'))
            nested=ul()
            with nested:
                for i in ['measures_table', 'roi_sizes_table','roi_coverage','measures_metadata']:
                    li(a(i.title(), href='#%s' % i))

    doc += hr()
    doc = create_section(doc, 'measureregistration', 'section', f"Registration")
    doc += hr()

    if mask:
        reg1_dict={}
        reg1_dict["image_list"] = [measure,mask]
        reg1_dict["image_png"] = os.path.join(image_dir,'mask_measure.png')
        reg1_dict["title"] = 'mask overlayed on measure'
        reg1_dict["divid"] = 'mask_measure'
        reg1_dict["divclass"] = 'img-caption'
        doc = AddRegSection(doc, reg1_dict,levellist=[0.5,5,10,15])
        doc += hr()

    reg1_dict={}
    reg1_dict["image_list"] = [measure,atlas]
    reg1_dict["image_png"] = os.path.join(image_dir,'atlas_measure.png')
    reg1_dict["title"] = 'atlas overlayed on measure'
    reg1_dict["divid"] = 'atlas_measure'
    reg1_dict["divclass"] = 'img-caption'
    doc = AddRegSection(doc, reg1_dict,levellist=[0.5,5,10,15],colors = ['g','b','y','m','c'],roi_index=0)
    doc += hr()

    if mask:
        reg1_dict={}
        reg1_dict["image_list"] = [measure,mask,atlas]
        reg1_dict["image_png"] = os.path.join(image_dir,'mask_atlas_measure.png')
        reg1_dict["title"] = 'mask overlaid on atlas and measure'
        reg1_dict["divid"] = 'mask_atlas_measure'
        reg1_dict["divclass"] = 'img-caption'
        doc = AddRegSection(doc, reg1_dict,levellist=[0.5,5,10,15],roi_index=1)
        doc += hr()


    doc += hr()
    doc = create_section(doc, 'measuresinfo', 'section', f"Measures Information")
    doc += hr()

    reg1_dict={}
    measures_table = metadata["FileCreated"]
    df = pd.read_table(measures_table,sep=",")
    reg1_dict["headers"] = df.columns.tolist()
    reg1_dict["rows"] = df.values.tolist()
    reg1_dict["title"] = "Measures"
    reg1_dict["divid"] = 'measures_table'
    reg1_dict["tabclass"] = 'data-table'
    reg1_dict["divclass"] = 'table-caption'
    doc = AddDataSection(doc, reg1_dict)
    doc += hr()

    reg1_dict={}
    roi_sizes_table = metadata["ROI Voxel Sizes"]
    df = pd.read_table(roi_sizes_table,sep=",")
    reg1_dict["headers"] = df.columns.tolist()
    reg1_dict["rows"] = df.values.tolist()
    reg1_dict["title"] = "ROI Voxel Sizes"
    reg1_dict["divid"] = 'roi_sizes_table'
    reg1_dict["tabclass"] = 'data-table'
    reg1_dict["divclass"] = 'table-caption'
    doc = AddDataSection(doc, reg1_dict)
    doc += hr()

    reg1_dict={}
    roi_sizes_table = metadata["ROI Coverage"]
    df = pd.read_table(roi_sizes_table,sep=",")
    reg1_dict["headers"] = df.columns.tolist()
    reg1_dict["rows"] = df.values.tolist()
    reg1_dict["title"] = "ROI Coverage"
    reg1_dict["divid"] = 'roi_coverage_table'
    reg1_dict["tabclass"] = 'data-table'
    reg1_dict["divclass"] = 'table-caption'
    doc = AddDataSection(doc, reg1_dict)
    doc += hr()

    reg1_dict={}
    reg1_dict["headers"] = ["Item", "Value"]
    reg1_dict["rows"] = [[k,metadata[k]] for k in metadata.keys()]
    reg1_dict["title"] = "Metadata"
    reg1_dict["divid"] = 'measures_metadata'
    reg1_dict["tabclass"] = 'info-table'
    reg1_dict["divclass"] = 'table-caption'
    doc = AddDataSection(doc, reg1_dict)
    doc += hr()

    return doc

def createRoiExtractReport(panpipe_labels,html_file, metadata,analysis_level=None):
    stylesheet = getParams(panpipe_labels,"STYLESHEET")
    pipeline = getParams(panpipe_labels,"PIPELINE")

    report_title = f"Html report for {pipeline}"
    if analysis_level == "participant":
        participant = getParams(panpipe_labels,"PARTICIPANT_LABEL")
        session = getParams(panpipe_labels,"PARTICIPANT_SESSION")
        report_title = f"Subject {participant}, Session {session} : " + report_title

    output_dir = os.path.dirname(html_file)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir,exist_ok=True)

    if isinstance(metadata,dict):
        metajson = metadata
    else:
        with open(metadata,"r") as infile:
            metajson = json.load(infile)

    image_dir=os.path.join(output_dir, 'images')

    doc_dict={}
    doc_dict["title"] = report_title
    doc_dict["stylesheet"] = f"{stylesheet}"
    doc_dict["image_dir"] = f"{image_dir}"
    doc_dict["output_dir"] = f"{output_dir}"
    if "Mask" in metajson.keys():
        doc_dict["mask"] = metajson["Mask"]
    else:
        doc_dict["mask"] = ""
    doc_dict["atlas"] = metajson["Atlas File"]
    doc_dict["measure"] = metajson["Input File"]
    doc_dict["pipeline"] = f"{pipeline}"
    doc_dict["metadata"] = metajson

    doc = addRoiExtractProcReport(doc_dict)

    html_file = newfile(assocfile=html_file,extension="html")
    html_file_working = newfile(assocfile=html_file,suffix="working")
    with open(html_file_working, 'w') as file:
        file.write(doc.render())
    make_html_images_inline(html_file_working, html_file)

    return html_file 


def addBasilMotionSection(doc,doc_dict):

    image_dir = doc_dict["image_dir"]
    debug_dir = doc_dict["debug_dir"] 
    fsl_tsplot=doc_dict["command_base"] + ' fsl_tsplot '

    #asldatapre=os.path.join(debug_dir,'asldata_orig.nii.gz')
    #image1=os.path.join(image_dir,'asldatapre.png')
    #asldatapost=os.path.join(debug_dir,'asldata.nii.gz')
    #image2=os.path.join(image_dir,'asldatapost.png')

    #reg1_dict={}
    #reg1_dict["image_list"] = [asldatapre]
    #reg1_dict["image_png"] = image1
    #reg1_dict["title"] = 'asldata before motion correction'
    #reg1_dict["divid"] = 'asldata_pre'
    #reg1_dict["divclass"] = 'img-caption'
    #doc = AddRegSection(doc, reg1_dict,levellist=[0.5,5,10,15])
    #doc += hr()

    #reg1_dict={}
    #reg1_dict["image_list"] = [asldatapost]
    #reg1_dict["image_png"] = image2
    #reg1_dict["divid"] = 'asldata_post'
    #reg1_dict["divclass"] = 'img-caption'
    #doc = AddRegSection(doc, reg1_dict,levellist=[0.5,5,10,15])
    #doc += hr()
    
    asldatapar=os.path.join(debug_dir,'asldata.par')
    image3=os.path.join(image_dir,'rot.png')
    fsl_tsplot_rot_command="{} -i {}  -t {} -u 1 --start=1 --finish=3 -a x,y,z -w 640 -h 144 -o {}".format(fsl_tsplot,asldatapar,'MCFLIRT_estimated_rotations_radians',image3)
    results = runCommand(fsl_tsplot_rot_command,UTLOGGER)
    doc = add_image(doc, "rotations", None, 'Rotation', image3)

    image4=os.path.join(image_dir,'trans.png')
    fsl_tsplot_trans_command="{} -i {}  -t {} -u 1 --start=4 --finish=6 -a x,y,z -w 640 -h 144 -o {}".format(fsl_tsplot,asldatapar,'MCFLIRT_estimated_translations_mm',image4)
    results = runCommand(fsl_tsplot_trans_command,UTLOGGER)
    doc = add_image(doc, "translations", None, 'Translation', image4)

    return doc


def addBasilStructRegSection(doc,doc_dict):

    title = doc_dict["title"]
    image_dir = doc_dict["image_dir"]
    debug_dir = doc_dict["debug_dir"] 

    struct=os.path.join(debug_dir,'struc.nii.gz')
    struct_bet_mask=os.path.join(debug_dir,'struc_bet_mask.nii.gz')
    image1=os.path.join(image_dir,'structmask_struct.png')
    calib=os.path.join(debug_dir,'calib_struc.nii.gz')
    image2=os.path.join(image_dir,'structmask_calib.png')
    asl=os.path.join(debug_dir,'asl2struct.nii.gz')
    image3=os.path.join(image_dir,'asl_struct.png')
    image4=os.path.join(image_dir,'asl_calib.png')

    reg1_dict={}
    reg1_dict["image_list"] = [struct,struct_bet_mask]
    reg1_dict["image_png"] = image1
    reg1_dict["title"] = 'brain mask overlaid on structural'
    reg1_dict["divid"] = 'structmask_struct'
    reg1_dict["divclass"] = 'img-caption'
    doc = AddRegSection(doc, reg1_dict,only_first = True,levellist=[0.5,5,10,15])
    doc += hr()

    reg1_dict={}
    reg1_dict["image_list"] = [calib,struct_bet_mask]
    reg1_dict["image_png"] = image2
    reg1_dict["title"] = 'brain mask overlaid on calibration image'
    reg1_dict["divid"] = 'structmask_calib'
    reg1_dict["divclass"] = 'img-caption'
    doc = AddRegSection(doc, reg1_dict,only_first = True,levellist=[0.5,5,10,15])
    doc += hr()

    reg1_dict={}
    reg1_dict["image_list"] = [struct,asl]
    reg1_dict["image_png"] = image3
    reg1_dict["title"] = 'asl overlaid on structural image'
    reg1_dict["divid"] = 'asl_struct'
    reg1_dict["divclass"] = 'img-caption'
    doc = AddRegSection(doc, reg1_dict,only_first = True,levellist=[0.5,5,10,15])
    doc += hr()

    reg1_dict={}
    reg1_dict["image_list"] = [calib,asl]
    reg1_dict["image_png"] = image4
    reg1_dict["title"] = 'asl overlaid on calib image'
    reg1_dict["divid"] = 'asl_calib'
    reg1_dict["divclass"] = 'img-caption'
    doc = AddRegSection(doc, reg1_dict,only_first = True,levellist=[0.5,5,10,15])
    doc += hr()

    return doc

def addBasilProcReport(doc_dict):
    stylesheet = doc_dict["stylesheet"]
    title = doc_dict["title"]

    doc = create_document(title, stylesheet)
    with doc:
        with div(id='links').add(ul()):
            h2('Contents')
            li(a('Structural Registration',href='#structuralreg'))
            nested=ul()
            with nested:
                for i in ['structmask_struct', 'structmask_calib','asl_struct','asl_calib']:
                    li(a(i.title(), href='#%s' % i))
            li(a('Motion Correction',href='#motioncorrection'))
            nested=ul()
            with nested:
                #for i in [ 'asldata_pre', 'asldata_post','rotations','translations']:
                for i in ['rotations','translations']:
                    li(a(i.title(), href='#%s' % i))

    doc += hr()
    doc = create_section(doc, 'structuralreg', None, 'Structural Registration')
    doc += hr()
    doc = addBasilStructRegSection(doc, doc_dict)
    doc += hr()
    doc = create_section(doc, 'motioncorrection', None, 'Motion Correction')
    doc += hr()
    doc = addBasilMotionSection(doc, doc_dict)

    return doc


def createBasilReport(panpipe_labels,html_file,debug_dir,analysis_level=None):
    stylesheet = getParams(panpipe_labels,"STYLESHEET")
    pipeline = getParams(panpipe_labels,"PIPELINE")
    report_title = f"Html report for {pipeline}"

    if analysis_level == "participant":
        participant = getParams(panpipe_labels,"PARTICIPANT_LABEL")
        session = getParams(panpipe_labels,"PARTICIPANT_SESSION")
        report_title = f"Subject {participant}, Session {session} : " + report_title

    output_dir = os.path.dirname(html_file)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir,exist_ok=True)

    image_dir=os.path.join(output_dir, 'images')

    doc_dict={}
    doc_dict["title"] = report_title
    doc_dict["stylesheet"] = f"{stylesheet}"
    doc_dict["image_dir"] = f"{image_dir}"
    doc_dict["output_dir"] = f"{output_dir}"
    doc_dict["debug_dir"] = debug_dir
    doc_dict["pipeline"] = f"{pipeline}"
    doc_dict["command_base"] = substitute_labels("<PAN_CONTAINER_RUN_OPTIONS> <PAN_CONTAINER> ",panpipe_labels)

    doc = addBasilProcReport(doc_dict)

    html_file = newfile(assocfile=html_file,extension="html")
    html_file_working = newfile(assocfile=html_file,suffix="working")
    with open(html_file_working, 'w') as file:
        file.write(doc.render())
    make_html_images_inline(html_file_working, html_file)

    return html_file 







from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
from nipype import logging as nlogging
import shutil

IFLOGGER=nlogging.getLogger('nipype.interface')

def sort_df(df,sort_dict):
    sort_columns=[]
    sort_order=[]
    if sort_dict:
        if isinstance(sort_dict,dict):
            for sort_col,sort_ord in sort_dict.items():
                if sort_col in df.columns:
                    sort_columns.append(sort_col)
                    sort_order.append(isTrue(sort_ord))
        elif isinstance(sort_dict,list):
            for sort_col in sort_dict:
                if sort_col in df.columns:
                    sort_columns.append(sort_col)
                    sort_order.append(True)
        else:
            if sort_dict in df.columns:
                sort_columns.append(sort_dict)
                sort_order.append(True)

    if sort_columns and sort_order:
        df = df.sort_values(by=sort_columns,ascending=sort_order)
        df = df.reset_index(drop=True)

    return df


def move_cols_front(df, cols):
    if not cols:
        cols = []
    cols_to_move=[]
    for col in cols:
        if col in df.columns:
            cols_to_move.append(col)
    if cols_to_move:
        return df[ cols_to_move + [ col for col in df.columns if col not in cols_to_move ] ]
    else:
        return df

def get_col_candidates(labels_dict,col_filter_key):
    filter_col_candidates=[]
    filter_file = getParams(labels_dict,col_filter_key)
    if filter_file:
        with open(filter_file, 'r') as infile:
            col_parts = infile.read()
            col_split_parts = col_parts.split("\n")
            filter_col_candidates = [x for x in col_split_parts if x]
    return filter_col_candidates


def filter_df(df,field_exceptions=[],filter_col_candidates=None,labels_dict={},col_filter_key="COL_FILTER_FILE"):

    current_cols = df.columns
    filter_cols = []

    if not filter_col_candidates:
        filter_col_candidates=[]
        if isinstance(col_filter_key,list):
            for key in col_filter_key:
                filter_col_candidates.extend(get_col_candidates(labels_dict,key))
        else:
            filter_col_candidates.extend(get_col_candidates(labels_dict,col_filter_key))

    if filter_col_candidates:
        filter_col_candidates.extend(field_exceptions)
        for col in filter_col_candidates:
            if col in current_cols:
                filter_cols.append(col)

        df=df[filter_cols]

    return df

def collate_csv_group_proc(labels_dict, csv_list1,csv_list2, add_prefix):

    if csv_list1 is None:
        csv_list1 = []
    if csv_list2 is None:
        csv_list2 = []
        
    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    output_dir=cwd
    participants_label = getParams(labels_dict,'GROUP_PARTICIPANTS_LABEL')
    participants_project = getParams(labels_dict,'GROUP_PARTICIPANTS_XNAT_PROJECT')
    participants_session = getParams(labels_dict,'GROUP_SESSION_LABEL')
    pipeline = getParams(labels_dict,'PIPELINE')

    participant_exclusions = getParams(labels_dict,"EXCLUDED_PARTICIPANTS")
    if participant_exclusions:
        subject_exclusions = process_exclusions(participant_exclusions)
    else:
        subject_exclusions = []

    ENABLE_SORT_COLUMNS = getParams(labels_dict,"ENABLE_SORT_COLUMNS")
    if ENABLE_SORT_COLUMNS:
        ENABLE_SORT_COLUMNS = isTrue(ENABLE_SORT_COLUMNS)
    else:
        ENABLE_SORT_COLUMNS=True

    RSEP = getParams(labels_dict,'RSEP')
    if not RSEP:
        RSEP=","
    LSEP = getParams(labels_dict,'LSEP')
    if not LSEP:
        LSEP=","

    csv_list_left=[]
    csv_list_right=[]

    EXPAND_CSV1=False
    if len(csv_list1)>0:
        for meas_template in csv_list1:
            if "<PARTICIPANT_LABEL>" in meas_template:
                EXPAND_CSV1=True
                break

    EXPAND_CSV2=False
    if len(csv_list2)>0:
        for meas_template in csv_list2:
            if "<PARTICIPANT_LABEL>" in meas_template:
                EXPAND_CSV2=True
                break

    if (participants_label is not None and (isinstance(participants_label,list) and len(participants_label) > 1)) and (participants_project is not None and (isinstance(participants_project,list) and len(participants_project)> 1)) and (EXPAND_CSV1 or EXPAND_CSV2):
        for part_vals in zip(participants_label,participants_project,participants_session):
            labels_dict = updateParams(labels_dict,"PARTICIPANT_LABEL",part_vals[0])
            labels_dict = updateParams(labels_dict,"PARTICIPANT_XNAT_PROJECT",part_vals[1])
            labels_dict = updateParams(labels_dict,"PARTICIPANT_SESSION",part_vals[2])
            for meas_template in csv_list1:
                evaluated_meas_template = substitute_labels(meas_template,labels_dict)
                csv_list_left.extend(glob.glob(evaluated_meas_template))
            for meas_template in csv_list2:
                evaluated_meas_template = substitute_labels(meas_template,labels_dict)
                csv_list_right.extend(glob.glob(evaluated_meas_template)) 
    elif (EXPAND_CSV1 or EXPAND_CSV2):
        labels_dict = updateParams(labels_dict,"PARTICIPANT_LABEL","*")
        labels_dict = updateParams(labels_dict,"PARTICIPANT_XNAT_PROJECT","*")
        labels_dict = updateParams(labels_dict,"PARTICIPANT_SESSION","*")
        for meas_template in csv_list1:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            csv_list_left.extend(glob.glob(evaluated_meas_template))
        for meas_template in csv_list2:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            csv_list_right.extend(glob.glob(evaluated_meas_template))
    else:
        for meas_template in csv_list1:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            csv_list_left.extend(glob.glob(evaluated_meas_template))
        for meas_template in csv_list2:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            csv_list_right.extend(glob.glob(evaluated_meas_template))


    csv_list_left = list(set(csv_list_left))
    csv_list_right = list(set(csv_list_right))

    csv_list_left.sort()
    csv_list_right.sort()
    
    roi_output_dir = os.path.join(cwd,'group_roi_output_dir')
    if not os.path.isdir(roi_output_dir):
        os.makedirs(roi_output_dir,exist_ok=True)

    roi_work_dir = os.path.join(cwd,'group_roi_work_dir')
    if not os.path.isdir(roi_work_dir):
        os.makedirs(roi_work_dir,exist_ok=True)

    IFLOGGER.info(f"List of csv files (left) to collate: {csv_list_left}")
    IFLOGGER.info(f"List of csv files (right) to collate: {csv_list_right}")

    out_files=[]

    roi_csv_inner = None
    cum_df_inner= None
    roi_csv_outer = None
    cum_df_outer= None

    roi_csv_inner_left = None
    roi_csv_outer_left = None
    cum_df_inner_left=pd.DataFrame()
    cum_df_outer_left=pd.DataFrame()

    collate_prefix_left = getParams(labels_dict,"COLLATE_PREFIX_LEFT")
    if not collate_prefix_left:
        collate_prefix_left = None

    collate_prefix_right = getParams(labels_dict,"COLLATE_PREFIX_RIGHT")
    if not collate_prefix_right:
        collate_prefix_right=None

    collate_join_left= getParams(labels_dict,"COLLATECOLS_JOIN_LEFT")
    if not collate_join_left:
        collate_join_left=["subject_id","session_id"]

    collate_join_right= getParams(labels_dict,"COLLATECOLS_JOIN_RIGHT")
    if not collate_join_right:
        collate_join_right=["subject_id","session_id"]
 
    len_csv_list_left = len(csv_list_left) 
    if len_csv_list_left > 0:
        for csv_file_index, csv_file in enumerate(csv_list_left):
            df = pd.read_table(csv_file,sep=LSEP)
            if cum_df_inner_left.empty:
                cum_df_inner_left = df
            elif csv_file_index == len_csv_list_left - 1:
                cum_df_inner_left = pd.concat([cum_df_inner_left,df],join="inner").drop_duplicates(ignore_index=True)
            else:
                cum_df_inner_left = pd.concat([cum_df_inner_left,df],join="inner")

            if cum_df_outer_left.empty:
                cum_df_outer_left = df
            elif csv_file_index == len_csv_list_left - 1:
                cum_df_outer_left = pd.concat([cum_df_outer_left,df],join="outer").drop_duplicates(ignore_index=True)
            else:
                cum_df_outer_left = pd.concat([cum_df_outer_left,df],join="outer")

        collate_name_left = getParams(labels_dict,"COLLATE_NAME_LEFT")
        if not collate_name_left:
            if not pipeline:
                collate_name_left="csvgroup"
            else:
                collate_name_left=f"{pipeline}"

        roi_csv_inner_left = os.path.join(roi_output_dir,'{}_{}_inner_left.csv'.format("group",collate_name_left))
        roi_csv_outer_left = os.path.join(roi_output_dir,'{}_{}_outer_left.csv'.format("group",collate_name_left))

        #create sorted output
        if ENABLE_SORT_COLUMNS:
            cum_df_inner_left = cum_df_inner_left.reindex(sorted(cum_df_inner_left.columns), axis=1)

        if collate_prefix_left:
            orig_cols = cum_df_inner_left.columns.tolist()
            other_cols = [x for x in orig_cols if x not in collate_join_left]
            new_cols = [f"{collate_prefix_left}.{x}" for x in other_cols]
            new_dict = dict(zip(other_cols,new_cols))
            cum_df_inner_left = cum_df_inner_left.rename(columns=new_dict)

        index_col_list = [collate_join_left.index(x) for x in cum_df_inner_left.columns if x in collate_join_left]
        index_col_list.sort()
        for index_col_num in index_col_list:
            index_col = cum_df_inner_left.pop(collate_join_left[index_col_num])
            cum_df_inner_left.insert(index_col_num,collate_join_left[index_col_num],index_col)


        #create sorted output
        if ENABLE_SORT_COLUMNS:
            cum_df_outer_left = cum_df_outer_left.reindex(sorted(cum_df_outer_left.columns), axis=1)
        if collate_prefix_left:
            orig_cols = cum_df_outer_left.columns.tolist()
            other_cols = [x for x in orig_cols if x not in collate_join_left]
            new_cols = [f"{collate_prefix_left}.{x}" for x in other_cols]
            new_dict = dict(zip(other_cols,new_cols))
            cum_df_outer_left = cum_df_outer_left.rename(columns=new_dict)

        index_col_list = [collate_join_left.index(x) for x in cum_df_outer_left.columns if x in collate_join_left]
        index_col_list.sort()
        for index_col_num in index_col_list:
            index_col = cum_df_outer_left.pop(collate_join_left[index_col_num])
            cum_df_outer_left.insert(index_col_num,collate_join_left[index_col_num],index_col)

        # deal with excluded
        cum_df_inner_left = mask_excludedrows(cum_df_inner_left, subject_exclusions, collate_join_left)
        work_csv_inner_left=newfile(outputdir=roi_work_dir,assocfile=roi_csv_inner_left)
        cum_df_inner_left.to_csv(work_csv_inner_left,sep=",",header=True, index=False)

        cum_df_outer_left = mask_excludedrows(cum_df_outer_left, subject_exclusions, collate_join_left)
        work_csv_outer_left=newfile(outputdir=roi_work_dir,assocfile=roi_csv_outer_left)
        cum_df_outer_left.to_csv(work_csv_outer_left,sep=",",header=True, index=False)

        # Drop or rename columns as specififed by LEFT_DROP and LEFT_RENAME
        left_inner_cols=list(cum_df_inner_left.columns.values)
        left_outer_cols=list(cum_df_outer_left.columns.values)
        LEFT_DROP = getParams(labels_dict,"LEFT_DROP")
        if LEFT_DROP:
            LEFT_COL_INNER=False
            LEFT_COL_OUTER=False
            if not isinstance(LEFT_DROP,list):
                LEFT_DROP=[LEFT_DROP]

            LEFT_DROP_CANDIDATES=[]
            for col_drop_candidate in LEFT_DROP:
                if "*" not in col_drop_candidate:
                    LEFT_DROP_CANDIDATES.append(col_drop_candidate)
                else:
                    left_drop_col_list = [ x for x in left_outer_cols if col_drop_candidate.replace("*","") in x]
                    LEFT_DROP_CANDIDATES.extend(left_drop_col_list)

            for col_drop in LEFT_DROP_CANDIDATES:
                if col_drop in left_inner_cols:
                    if col_drop in collate_join_left:
                        IFLOGGER.debug(f"Cannot drop {col_drop} as it is a join column in {collate_join_left}")
                    else:
                        IFLOGGER.debug(f"about to drop {col_drop} from left_inner_cols")
                        left_inner_cols.pop(left_inner_cols.index(col_drop))
                        LEFT_COL_INNER = True

                if col_drop in left_outer_cols:
                    if col_drop in collate_join_left:
                        IFLOGGER.debug(f"Cannot drop {col_drop} as it is a join column in {collate_join_left}")
                    else:
                        IFLOGGER.debug(f"about to drop {col_drop} from left_outer_cols")
                        left_outer_cols.pop(left_outer_cols.index(col_drop))
                        LEFT_COL_OUTER = True

            if LEFT_COL_INNER:
                cum_df_inner_left=cum_df_inner_left[left_inner_cols]
                IFLOGGER.debug(f"Created cum_df_inner_left with columns {left_inner_cols}")

            if LEFT_COL_OUTER:
                cum_df_outer_left=cum_df_outer_left[left_outer_cols]
                IFLOGGER.debug(f"Created cum_df_outer_left with columns {left_outer_cols}")

        left_inner_cols=list(cum_df_inner_left.columns.values)
        left_outer_cols=list(cum_df_outer_left.columns.values)
        LEFT_RENAME = getParams(labels_dict,"LEFT_RENAME")
        if LEFT_RENAME:
            left_inner_dict={}
            left_outer_dict={}
            if isinstance(LEFT_RENAME,dict):
                for itemkey,itemvalue in LEFT_RENAME.items():
                    if itemkey in left_inner_cols:
                        if itemkey in collate_join_left:
                            IFLOGGER.debug(f"Cannot rename {itemkey} as it is a join column in {collate_join_left}")
                        else:
                            left_inner_dict[itemkey]=itemvalue
                            left_inner_cols.append(itemvalue)
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in left_inner_cols {left_inner_cols}. Skipping this rename {itemkey}:{itemvalue}")

                    if itemkey in left_outer_cols:
                        if itemkey in collate_join_left:
                            IFLOGGER.debug(f"Cannot rename {itemkey} as it is a join column in {collate_join_left}")
                        else:
                            left_outer_dict[itemkey]=itemvalue
                            left_outer_cols.append(itemvalue)
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in left_inner_cols {left_outer_cols}. Skipping this rename {itemkey}:{itemvalue}")

                if left_inner_dict:
                    cum_df_inner_left = cum_df_inner_left.rename(columns=left_inner_dict)

                if left_outer_dict:
                    cum_df_outer_left = cum_df_outer_left.rename(columns=left_outer_dict)
            else:
                IFLOGGER.debug(f"LEFT_RENAME should be a dictionary of values but {LEFT_RENAME} passed")

        left_inner_cols=list(cum_df_inner_left.columns.values)
        left_outer_cols=list(cum_df_outer_left.columns.values)
        LEFT_COPY = getParams(labels_dict,"LEFT_COPY")
        if LEFT_COPY:
            if isinstance(LEFT_COPY,dict):
                for itemkey,itemvalue in LEFT_COPY.items():
                    if itemkey == itemvalue:
                        IFLOGGER.debug(f"Column {itemkey}to copy from is the same as column to copy to {itemvalue}. Skipping this rename {itemkey}:{itemvalue}")
                    else:
                        if itemkey in left_inner_cols:
                            cum_df_inner_left[itemvalue]=cum_df_inner_left[itemkey]
                            left_inner_cols.append(itemvalue)
                        else:
                            IFLOGGER.debug(f"Column {itemkey} not found in left_inner_cols {left_inner_cols}. Skipping this copy {itemkey}:{itemvalue}")

                        if itemkey in left_outer_cols:
                            cum_df_outer_left[itemvalue]=cum_df_outer_left[itemkey]
                            left_outer_cols.append(itemvalue)
                        else:
                            IFLOGGER.debug(f"Column {itemkey} not found in left_outer_cols {left_outer_cols}. Skipping this copy {itemkey}:{itemvalue}")

            else:
                IFLOGGER.debug(f"LEFT_COPY should be a dictionary of values but {LEFT_COPY} passed")

        left_inner_cols=list(cum_df_inner_left.columns.values)
        left_outer_cols=list(cum_df_outer_left.columns.values)
        LEFT_TRANSLATE = getParams(labels_dict,"LEFT_TRANSLATE")
        if LEFT_TRANSLATE:
            if isinstance(LEFT_TRANSLATE,dict):
                for itemkey,itemvalue in LEFT_TRANSLATE.items():
                    if itemkey in left_inner_cols:
                        cum_df_inner_left[itemkey] = cum_df_inner_left[itemkey].apply(lambda row :  row_translate(row,itemvalue))
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in left_inner_cols {left_inner_cols}. Skipping this translation {itemkey}:{itemvalue}")

                    if itemkey in left_outer_cols:
                        cum_df_outer_left[itemkey] = cum_df_outer_left[itemkey].apply(lambda row :  row_translate(row,itemvalue))
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in left_outer_cols {left_outer_cols}. Skipping this translation {itemkey}:{itemvalue}")

            else:
                IFLOGGER.debug(f"LEFT_TRANSLATE should be a dictionary of values but {LEFT_TRANSLATE} passed")          


    roi_csv_inner_right = None
    roi_csv_outer_right = None
    cum_df_inner_right=pd.DataFrame()
    cum_df_outer_right=pd.DataFrame()

    len_csv_list_right = len(csv_list_right)
    if  len_csv_list_right > 0:
        for csv_file_index, csv_file in enumerate(csv_list_right):
            df = pd.read_table(csv_file,sep=RSEP)
            if cum_df_inner_right.empty:
                cum_df_inner_right = df
            elif csv_file_index == len_csv_list_right - 1:
                cum_df_inner_right = pd.concat([cum_df_inner_right,df],join="inner").drop_duplicates(ignore_index=True)
            else:
                cum_df_inner_right = pd.concat([cum_df_inner_right,df],join="inner")

            if cum_df_outer_right.empty:
                cum_df_outer_right = df
            elif csv_file_index == len_csv_list_right - 1:
                cum_df_outer_right = pd.concat([cum_df_outer_right,df],join="outer").drop_duplicates(ignore_index=True)
            else:
                cum_df_outer_right = pd.concat([cum_df_outer_right,df],join="outer")

        collate_name_right = getParams(labels_dict,"COLLATE_NAME_RIGHT")
        if not collate_name_right:
            if not pipeline:
                collate_name_right="csvgroup"
            else:
                collate_name_right="pipeline"


        roi_csv_inner_right = os.path.join(roi_output_dir,'{}_{}_inner_right.csv'.format("group",collate_name_right))
        roi_csv_outer_right = os.path.join(roi_output_dir,'{}_{}_outer_right.csv'.format("group",collate_name_right))

        #create sorted output
        if ENABLE_SORT_COLUMNS:
            cum_df_inner_right = cum_df_inner_right.reindex(sorted(cum_df_inner_right.columns), axis=1)
        if collate_prefix_right:
            orig_cols = cum_df_inner_right.columns.tolist()
            other_cols = [x for x in orig_cols if x not in collate_join_right]
            new_cols = [f"{collate_prefix_right}.{x}" for x in other_cols]
            new_dict = dict(zip(other_cols,new_cols))
            cum_df_inner_right = cum_df_inner_right.rename(columns=new_dict)

        index_col_list = [collate_join_right.index(x) for x in cum_df_inner_right.columns if x in collate_join_right]
        index_col_list.sort()
        for index_col_num in index_col_list:
            index_col = cum_df_inner_right.pop(collate_join_right[index_col_num])
            cum_df_inner_right.insert(index_col_num,collate_join_right[index_col_num],index_col)

        #create sorted output
        if ENABLE_SORT_COLUMNS:
            cum_df_outer_right = cum_df_outer_right.reindex(sorted(cum_df_outer_right.columns), axis=1)
        if collate_prefix_right:
            orig_cols = cum_df_outer_right.columns.tolist()
            other_cols = [x for x in orig_cols if x not in collate_join_right]
            new_cols = [f"{collate_prefix_right}.{x}" for x in other_cols]
            new_dict = dict(zip(other_cols,new_cols))
            cum_df_outer_right = cum_df_outer_right.rename(columns=new_dict)

        index_col_list = [collate_join_right.index(x) for x in cum_df_outer_right.columns if x in collate_join_right]
        index_col_list.sort()
        for index_col_num in index_col_list:
            index_col = cum_df_outer_right.pop(collate_join_right[index_col_num])
            cum_df_outer_right.insert(index_col_num,collate_join_right[index_col_num],index_col)


        cum_df_inner_right = mask_excludedrows(cum_df_inner_right, subject_exclusions, collate_join_right) 
        work_csv_inner_right=newfile(outputdir=roi_work_dir,assocfile=roi_csv_inner_right)    
        cum_df_inner_right.to_csv(work_csv_inner_right,sep=",",header=True, index=False)

        cum_df_outer_right = mask_excludedrows(cum_df_outer_right, subject_exclusions, collate_join_right)
        work_csv_outer_right=newfile(outputdir=roi_work_dir,assocfile=roi_csv_outer_right)    
        cum_df_outer_right.to_csv(work_csv_outer_right,sep=",",header=True, index=False)

        # Drop or rename columns as specififed by RIGHT_DROP and RIGHT_RENAME
        right_inner_cols=list(cum_df_inner_right.columns.values)
        right_outer_cols=list(cum_df_outer_right.columns.values)
        RIGHT_DROP = getParams(labels_dict,"RIGHT_DROP")
        if RIGHT_DROP:
            RIGHT_COL_INNER=False
            RIGHT_COL_OUTER=False
            if not isinstance(RIGHT_DROP,list):
                RIGHT_DROP=[RIGHT_DROP]

            RIGHT_DROP_CANDIDATES=[]
            for col_drop_candidate in RIGHT_DROP:
                if "*" not in col_drop_candidate:
                    RIGHT_DROP_CANDIDATES.append(col_drop_candidate)
                else:
                    right_drop_col_list = [ x for x in right_outer_cols if col_drop_candidate.replace("*","") in x]
                    RIGHT_DROP_CANDIDATES.extend(right_drop_col_list)

            for col_drop in RIGHT_DROP_CANDIDATES:
                if col_drop in right_inner_cols:
                    if col_drop in collate_join_right:
                        IFLOGGER.debug(f"Cannot drop {col_drop} as it is a join column in {collate_join_right}")
                    else:
                        IFLOGGER.debug(f"about to drop {col_drop} from right_inner_cols")
                        right_inner_cols.pop(right_inner_cols.index(col_drop))
                        RIGHT_COL_INNER = True

                if col_drop in right_outer_cols:
                    if col_drop in collate_join_right:
                        IFLOGGER.debug(f"Cannot drop {col_drop} as it is a join column in {collate_join_right}")
                    else:
                        IFLOGGER.debug(f"about to drop {col_drop} from right_outer_cols")
                        right_outer_cols.pop(right_outer_cols.index(col_drop))
                        RIGHT_COL_OUTER = True

            if RIGHT_COL_INNER:
                cum_df_inner_right=cum_df_inner_right[right_inner_cols]
                IFLOGGER.debug(f"Created cum_df_inner_right with columns {right_inner_cols}")

            if RIGHT_COL_OUTER:
                cum_df_outer_right=cum_df_outer_right[right_outer_cols]
                IFLOGGER.debug(f"Created cum_df_outer_right with columns {right_outer_cols}")

        right_inner_cols=list(cum_df_inner_right.columns.values)
        right_outer_cols=list(cum_df_outer_right.columns.values)
        RIGHT_RENAME = getParams(labels_dict,"RIGHT_RENAME")
        if RIGHT_RENAME:
            right_inner_dict={}
            right_outer_dict={}
            if isinstance(RIGHT_RENAME,dict):
                for itemkey,itemvalue in RIGHT_RENAME.items():
                    if itemkey in right_inner_cols:
                        if itemkey in collate_join_right:
                            IFLOGGER.debug(f"Cannot rename {itemkey} as it is a join column in {collate_join_right}")
                        else:
                            right_inner_dict[itemkey]=itemvalue
                            right_inner_cols.append(itemvalue)
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in right_inner_cols {right_inner_cols}. Skipping this rename {itemkey}:{itemvalue}")

                    if itemkey in right_outer_cols:
                        if itemkey in collate_join_right:
                            IFLOGGER.debug(f"Cannot rename {itemkey} as it is a join column in {collate_join_right}")
                        else:
                            right_outer_dict[itemkey]=itemvalue
                            right_outer_cols.append(itemvalue)
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in right_inner_cols {right_outer_cols}. Skipping this rename {itemkey}:{itemvalue}")

                if right_inner_dict:
                    cum_df_inner_right = cum_df_inner_right.rename(columns=right_inner_dict)

                if right_outer_dict:
                    cum_df_outer_right = cum_df_outer_right.rename(columns=right_outer_dict)


            else:
                IFLOGGER.debug(f"RIGHT_RENAME should be a dictionary of values but {RIGHT_RENAME} passed")

        right_inner_cols=list(cum_df_inner_right.columns.values)
        right_outer_cols=list(cum_df_outer_right.columns.values)
        RIGHT_COPY = getParams(labels_dict,"RIGHT_COPY")
        if RIGHT_COPY:
            if isinstance(RIGHT_COPY,dict):
                for itemkey,itemvalue in RIGHT_TRANSLATE.items():
                    if itemkey == itemvalue:
                        IFLOGGER.debug(f"Column {itemkey}to copy from is the same as column to copy to {itemvalue}. Skipping this rename {itemkey}:{itemvalue}")
                    else:
                        if itemkey in right_inner_cols:
                            cum_df_inner_right[itemvalue]=cum_df_inner_right[itemkey]
                            right_inner_cols.append(itemvalue)
                        else:
                            IFLOGGER.debug(f"Column {itemkey} not found in right_inner_cols {right_inner_cols}. Skipping this copy {itemkey}:{itemvalue}")

                        if itemkey in right_outer_cols:
                            cum_df_outer_right[itemvalue]=cum_df_outer_right[itemkey]
                            right_outer_cols.append(itemvalue)
                        else:
                            IFLOGGER.debug(f"Column {itemkey} not found in right_outer_cols {right_outer_cols}. Skipping this copy {itemkey}:{itemvalue}")

            else:
                IFLOGGER.debug(f"RIGHT_COPY should be a dictionary of values but {RIGHT_COPY} passed")


        right_inner_cols=list(cum_df_inner_right.columns.values)
        right_outer_cols=list(cum_df_outer_right.columns.values)
        RIGHT_TRANSLATE = getParams(labels_dict,"RIGHT_TRANSLATE")
        if RIGHT_TRANSLATE:
            if isinstance(RIGHT_TRANSLATE,dict):
                for itemkey,itemvalue in RIGHT_COPY.items():
                    if itemkey in right_inner_cols:
                        cum_df_inner_right[itemkey] = cum_df_inner_right[itemkey].apply(lambda row :  row_translate(row,itemvalue))
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in right_inner_cols {right_inner_cols}. Skipping this translation {itemkey}:{itemvalue}")

                    if itemkey in right_outer_cols:
                        cum_df_outer_right[itemkey] = cum_df_outer_right[itemkey].apply(lambda row :  row_translate(row,itemvalue))
                    else:
                        IFLOGGER.debug(f"Column {itemkey} not found in right_outer_cols {right_outer_cols}. Skipping this translation {itemkey}:{itemvalue}")

            else:
                IFLOGGER.debug(f"RIGHT_TRANSLATE should be a dictionary of values but {RIGHT_TRANSLATE} passed")   


    ALL_GROUP = isTrue(getParams(labels_dict,"ALL_GROUP"))
    # this flag currently also controls if cumulative group results are appended - need a better way to do this
    FILTER_GROUP = isTrue(getParams(labels_dict,"FILTER_GROUP"))
    LAST_OUTPUT_FILES = getParams(labels_dict,"LAST_OUTPUT_FILES")
    ADD_CUMULATIVE = getParams(labels_dict,"ADD_CUMULATIVE")

    SORT_VALUES = getParams(labels_dict,"SORT_VALUES")

    filter_set = None
    if not ALL_GROUP and FILTER_GROUP:
        IFLOGGER.info("Group Filter: Attempting to filter out specific participant and session from dataframe using subject_id and session_id")
        filter_set = set(zip([f"sub-{x}" for x in participants_label],[f"ses-{x}" for x in participants_session]))

    # replace hyphens with underscores
    if not cum_df_inner_right.empty:
        cum_df_inner_right.columns = cum_df_inner_right.columns.str.replace("-", "_", regex=True)
        if filter_set:
            cum_df_inner_right = cum_df_inner_right[cum_df_inner_right[['subject_id', 'session_id']].apply(tuple, axis=1).isin(filter_set)]
        extra_field_exceptions = getParams(labels_dict,"RIGHT_FILTER_EXCEPTIONS")
        if not extra_field_exceptions:
            extra_field_exceptions = []
        cum_df_inner_right = filter_df(cum_df_inner_right,field_exceptions=collate_join_right + extra_field_exceptions,labels_dict=labels_dict,col_filter_key="RIGHT_FILTER_FILE")
        cum_df_inner_right = move_cols_front(cum_df_inner_right, getParams(labels_dict,"COLUMNS_TO_FRONT"))
        cum_df_inner_right = sort_df(cum_df_inner_right,SORT_VALUES)
        cum_df_inner_right.to_csv(roi_csv_inner_right,sep=",",header=True, index=False)
    
    if not cum_df_inner_left.empty:
        cum_df_inner_left.columns = cum_df_inner_left.columns.str.replace("-", "_", regex=True)
        if filter_set:
            cum_df_inner_left = cum_df_inner_left[cum_df_inner_left[['subject_id', 'session_id']].apply(tuple, axis=1).isin(filter_set)]
        extra_field_exceptions = getParams(labels_dict,"LEFT_FILTER_EXCEPTIONS")
        if not extra_field_exceptions:
            extra_field_exceptions = []
        cum_df_inner_left = filter_df(cum_df_inner_left,field_exceptions=collate_join_left + extra_field_exceptions,labels_dict=labels_dict,col_filter_key="LEFT_FILTER_FILE")
        cum_df_inner_left = move_cols_front(cum_df_inner_left, getParams(labels_dict,"COLUMNS_TO_FRONT"))
        cum_df_inner_left = sort_df(cum_df_inner_left,SORT_VALUES)
        cum_df_inner_left.to_csv(roi_csv_inner_left,sep=",",header=True, index=False)

    if not cum_df_outer_right.empty:
        cum_df_outer_right.columns = cum_df_outer_right.columns.str.replace("-", "_", regex=True)
        if filter_set:
            cum_df_outer_right = cum_df_outer_right[cum_df_outer_right[['subject_id', 'session_id']].apply(tuple, axis=1).isin(filter_set)]
        extra_field_exceptions = getParams(labels_dict,"RIGHT_FILTER_EXCEPTIONS")
        if not extra_field_exceptions:
            extra_field_exceptions = []
        cum_df_outer_right = filter_df(cum_df_outer_right,field_exceptions=collate_join_right + extra_field_exceptions,labels_dict=labels_dict,col_filter_key="RIGHT_FILTER_FILE")
        cum_df_outer_right = move_cols_front(cum_df_outer_right, getParams(labels_dict,"COLUMNS_TO_FRONT"))
        cum_df_outer_right = sort_df(cum_df_outer_right,SORT_VALUES)
        cum_df_outer_right.to_csv(roi_csv_outer_right,sep=",",header=True, index=False)
    
    if not cum_df_outer_left.empty:
        cum_df_outer_left.columns = cum_df_outer_left.columns.str.replace("-", "_", regex=True)
        if filter_set:
            cum_df_outer_left = cum_df_outer_left[cum_df_outer_left[['subject_id', 'session_id']].apply(tuple, axis=1).isin(filter_set)]
        extra_field_exceptions = getParams(labels_dict,"LEFT_FILTER_EXCEPTIONS")
        if not extra_field_exceptions:
            extra_field_exceptions = []
        cum_df_outer_left = filter_df(cum_df_outer_left,field_exceptions=collate_join_left + extra_field_exceptions,labels_dict=labels_dict,col_filter_key="LEFT_FILTER_FILE")
        cum_df_outer_left = move_cols_front(cum_df_outer_left, getParams(labels_dict,"COLUMNS_TO_FRONT"))
        cum_df_outer_left = sort_df(cum_df_outer_left,SORT_VALUES)
        cum_df_outer_left.to_csv(roi_csv_outer_left,sep=",",header=True, index=False)


    if not cum_df_inner_right.empty and not cum_df_inner_left.empty:

        # if a column is duplicated in left and right then quick hack is to just remove it from the right table, except for the join columns
        left_inner_cols=list(cum_df_inner_left.columns.values)
        right_inner_cols=list(cum_df_inner_right.columns.values)
        right_inner_cols_to_drop = set(left_inner_cols).intersection(set(right_inner_cols)).difference(set(collate_join_right))
        IFLOGGER.info(f"Dropping duplicated columns {right_inner_cols_to_drop} from right table")
        cum_df_inner_right = cum_df_inner_right.drop(right_inner_cols_to_drop, axis=1)

        cum_df_inner = pd.merge(cum_df_inner_left, cum_df_inner_right,  how='left', left_on=collate_join_left, right_on =collate_join_right)
        roi_csv_inner = os.path.join(roi_output_dir,'{}_{}-{}_inner.csv'.format("final-group",collate_name_left,collate_name_right))

        #create sorted output
        if ENABLE_SORT_COLUMNS:
            cum_df_inner = cum_df_inner.reindex(sorted(cum_df_inner.columns), axis=1)
        index_col_list = [collate_join_left.index(x) for x in cum_df_inner.columns if x in collate_join_left]
        index_col_list.sort()
        for index_col_num in index_col_list:
            index_col = cum_df_inner.pop(collate_join_left[index_col_num])
            cum_df_inner.insert(index_col_num,collate_join_left[index_col_num],index_col)

        cum_df_inner = move_cols_front(cum_df_inner, getParams(labels_dict,"COLUMNS_TO_FRONT"))
        cum_df_inner.to_csv(roi_csv_inner,sep=",",header=True, index=False)

        metadata = {}
        created_datetime = get_datetimestring_utc()
        metadata["History"]={}
        metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: Inner Join (both tables)")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Only matched columns are retained.")
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{roi_csv_inner_left},{roi_csv_inner_right}")
        roi_csv_inner_json = create_metadata(roi_csv_inner, created_datetime, metadata = metadata)

    elif not cum_df_inner_right.empty:
        roi_csv_inner = roi_csv_inner_right
        metadata = {}
        created_datetime = get_datetimestring_utc()
        metadata["History"]={}
        metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: collate single table without join")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Only matched columns are retained.")
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{roi_csv_inner_left},{roi_csv_inner_right}")
        roi_csv_inner_json = create_metadata(roi_csv_inner, created_datetime, metadata = metadata)

    elif not cum_df_inner_left.empty:
        roi_csv_inner = roi_csv_inner_left
        metadata = {}
        created_datetime = get_datetimestring_utc()
        metadata["History"]={}
        metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: collate single table without join")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Only matched columns are retained.")
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{roi_csv_inner_left},{roi_csv_inner_right}")
        roi_csv_inner_json = create_metadata(roi_csv_inner, created_datetime, metadata = metadata)

    if not cum_df_outer_right.empty and not cum_df_outer_left.empty:

        # if a column is duplicated in left and right then quick hack is to just remove it from the right table, except for the join columns
        left_outer_cols=list(cum_df_outer_left.columns.values)
        right_outer_cols=list(cum_df_outer_right.columns.values)
        right_outer_cols_to_drop = set(left_outer_cols).intersection(set(right_outer_cols)).difference(set(collate_join_right))
        IFLOGGER.info(f"Dropping duplicated columns {right_outer_cols_to_drop} from right table")
        cum_df_outer_right = cum_df_outer_right.drop(right_outer_cols_to_drop, axis=1)

        cum_df_outer = pd.merge(cum_df_outer_left, cum_df_outer_right,  how='left', left_on=collate_join_left, right_on =collate_join_right)
        roi_csv_outer = os.path.join(roi_output_dir,'{}_{}-{}_outer.csv'.format("final-group",collate_name_left,collate_name_right))

        #create sorted output
        if ENABLE_SORT_COLUMNS:
            cum_df_outer = cum_df_outer.reindex(sorted(cum_df_outer.columns), axis=1)

        index_col_list = [collate_join_left.index(x) for x in cum_df_outer.columns if x in collate_join_left]
        index_col_list.sort()
        for index_col_num in index_col_list:
            index_col = cum_df_outer.pop(collate_join_left[index_col_num])
            cum_df_outer.insert(index_col_num,collate_join_left[index_col_num],index_col)

        cum_df_outer = move_cols_front(cum_df_outer, getParams(labels_dict,"COLUMNS_TO_FRONT"))
        cum_df_outer.to_csv(roi_csv_outer,sep=",",header=True, index=False)
        metadata = {}
        created_datetime = get_datetimestring_utc()
        metadata["History"]={}
        metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: Outer Join (both tables)")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Unmatched columns are retained.")
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{roi_csv_outer_left},{roi_csv_outer_right}")
        roi_csv_outer_json = create_metadata(roi_csv_outer, created_datetime, metadata = metadata)


    elif not cum_df_outer_right.empty:
        roi_csv_outer = roi_csv_outer_right
        metadata = {}
        created_datetime = get_datetimestring_utc()
        metadata["History"]={}
        metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: collate single table without join")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Unmatched columns are retained.")
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{roi_csv_outer}")
        roi_csv_outer_json = create_metadata(roi_csv_outer, created_datetime, metadata = metadata)
    elif not cum_df_outer_left.empty:
        roi_csv_outer = roi_csv_outer_left
        metadata = {}
        created_datetime = get_datetimestring_utc()
        metadata["History"]={}
        metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: collate single table without join")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Unmatched columns are retained.")
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{roi_csv_outer}")
        roi_csv_outer_json = create_metadata(roi_csv_outer, created_datetime, metadata = metadata)

    # this workflow needs to be looked at !!!!
    if not ALL_GROUP and LAST_OUTPUT_FILES and not FILTER_GROUP and ADD_CUMULATIVE:
        IFLOGGER.info("This is an update to the last group files. We will append results from this run to files at {LAST_OUTPUT_FILES}")
        for group_file in LAST_OUTPUT_FILES:

            if os.path.basename(group_file) == os.path.basename(roi_csv_inner):
                new_cum_df_inner = pd.DataFrame()
                group_df = pd.read_table(group_file,sep=LSEP)
                new_df = pd.read_table(roi_csv_inner,sep=LSEP)
                new_cum_df_inner = pd.concat([group_df,new_df],join="inner").drop_duplicates(ignore_index=True)

                if new_cum_df_inner.empty:
                    IFLOGGER.info("Problem reading files {group_file} or {roi_csv_inner}.")
                else:
                    sorted_df = new_cum_df_inner.sort_values(by = collate_join_left, ascending = [True for x in range(len(collate_join_left))])
                    sorted_df.reset_index(drop=True,inplace=True)
                    created_datetime = get_datetimestring_utc()
                    roi_csv_inner_backup = newfile(assocfile=roi_csv_inner,suffix=f"{created_datetime}")
                    shutil.copy(roi_csv_inner,roi_csv_inner_backup)
                    sorted_df = filter_df(sorted_df,field_exceptions=collate_join_right + collate_join_left,labels_dict=labels_dict,col_filter_key=["LEFT_FILTER_FILE","RIGHT_FILTER_FILE"])
                    sorted_df.to_csv(roi_csv_inner,sep=LSEP,header=True, index=False)

                    roi_csv_inner_json_backup = newfile(assocfile=roi_csv_inner_json,suffix=f"{created_datetime}")
                    shutil.copy(roi_csv_inner_json,roi_csv_inner_json_backup)

                    metadata = {}
                    metadata["History"]={}
                    old_metadata = {}
                    previous_metadata_file = newfile(assocfile=group_file,extension="json")
                    with open(previous_metadata_file,"r") as infile:
                        old_metadata = json.load(infile)
                    if old_metadata and "History" in old_metadata.keys():
                        if isinstance(old_metadata["History"],dict):
                            for itemkey,itemvalue in old_metadata["History"].items():
                                metadata["History"][itemkey] = itemvalue
                        else:
                            metadata["History"]["UnknownDate"] = old_metadata["History"]


                    created_datetime = get_datetimestring_utc()
                    metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
                    metadata = updateParams(metadata,"Title","collate_csv_group.py: Inner Join (both tables)")
                    metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Only matched columns are retained.")
                    metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
                    metadata = updateParams(metadata,"InputFiles",f"{roi_csv_inner_left},{roi_csv_inner_right}")
                    roi_csv_inner_json = create_metadata(roi_csv_inner, created_datetime, metadata = metadata)



            if os.path.basename(group_file) == os.path.basename(roi_csv_outer):
                new_cum_df_outer  = pd.DataFrame()
                group_df = pd.read_table(group_file,sep=LSEP)
                new_df = pd.read_table(roi_csv_outer,sep=LSEP)
                new_cum_df_outer = pd.concat([group_df,new_df],join="outer").drop_duplicates(ignore_index=True)

                if new_cum_df_outer.empty:
                    IFLOGGER.info("Problem reading files {group_file} or {roi_csv_outer}.")
                else:
                    sorted_df = new_cum_df_outer.sort_values(by = collate_join_left, ascending = [True for x in range(len(collate_join_left))])
                    sorted_df.reset_index(drop=True,inplace=True)
                    created_datetime = get_datetimestring_utc()
                    roi_csv_outer_backup = newfile(assocfile=roi_csv_outer,suffix=f"{created_datetime}")
                    shutil.copy(roi_csv_outer,roi_csv_outer_backup)
                    sorted_df = filter_df(sorted_df,field_exceptions=collate_join_right + collate_join_left,labels_dict=labels_dict,col_filter_key=["LEFT_FILTER_FILE","RIGHT_FILTER_FILE"])
                    sorted_df.to_csv(roi_csv_outer,sep=LSEP,header=True, index=False)

                    roi_csv_outer_json_backup = newfile(assocfile=roi_csv_outer_json,suffix=f"{created_datetime}")
                    shutil.copy(roi_csv_outer_json,roi_csv_outer_json_backup)


                    metadata = {}
                    metadata["History"]={}
                    old_metadata = {}
                    previous_metadata_file = newfile(assocfile=group_file,extension="json")
                    with open(previous_metadata_file,"r") as infile:
                        old_metadata = json.load(infile)
                    if old_metadata and "History" in old_metadata.keys():
                        if isinstance(old_metadata["History"],dict):
                            for itemkey,itemvalue in old_metadata["History"].items():
                                metadata["History"][itemkey] = itemvalue
                        else:
                            metadata["History"]["UnknownDate"] = old_metadata["History"]

                    created_datetime = get_datetimestring_utc()
                    metadata["History"][created_datetime] = f"Updated with {len(participants_label)} subjects from {participants_label[0]} to {participants_label[-1]}"
                    metadata = updateParams(metadata,"Title","collate_csv_group.py: collate single table without join")
                    metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Unmatched columns are retained.")
                    metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
                    metadata = updateParams(metadata,"InputFiles",f"{roi_csv_outer}")
                    roi_csv_outer_json = create_metadata(roi_csv_outer, created_datetime, metadata = metadata)


    out_files.insert(0,roi_csv_inner)
    out_files.insert(1,roi_csv_inner_json)
    out_files.insert(2,roi_csv_outer)
    out_files.insert(3,roi_csv_outer_json)

    return {
        "roi_csv_inner":roi_csv_inner,
        "roi_csv_inner_metadata":roi_csv_inner_json,
        "roi_csv_outer":roi_csv_outer,
        "roi_csv_outer_metadata":roi_csv_outer_json,
        "roi_output_dir":roi_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }


class collate_csv_groupInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    csv_list1 = traits.List(desc='list of files')
    csv_list2 = traits.List(desc='list of files')
    add_prefix = traits.Bool(False,desc="Create header prefix while joining tables",usedefault=True)

class collate_csv_groupOutputSpec(TraitedSpec):
    roi_csv_inner = File(desc='CSV file of results')
    roi_csv_inner_metadata = File(desc='metadata for CSV results file - inner join')
    roi_csv_outer = File(desc='CSV file of results')
    roi_csv_outer_metadata = File(desc='metadata for CSV results file - outer join')
    roi_output_dir = traits.String(desc='roi output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class collate_csv_group_pan(BaseInterface):
    input_spec = collate_csv_groupInputSpec
    output_spec = collate_csv_groupOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = collate_csv_group_proc(
            self.inputs.labels_dict,
            self.inputs.csv_list1,
            self.inputs.csv_list2,
            self.inputs.add_prefix
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="collate_csv_group_node",csv_list1="",csv_list2="",add_prefix=False,LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(collate_csv_group_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if not csv_list1 is None and not csv_list1 == "":
        pan_node.inputs.csv_list1 = csv_list1
 
    if not csv_list2 is None and not csv_list2 == "":
        pan_node.inputs.csv_list2 = csv_list2

    pan_node.inputs.add_prefix =  add_prefix

    return pan_node





from sqlalchemy import create_engine, text
import pandas as pd
import json
import argparse
import os
from urllib.parse import quote_plus
from pathlib import Path
from panpipelines.utils.util_functions import *


# Set up argument parser
def parse_params():
    parser = argparse.ArgumentParser(description="Execute an SQL query and export the results to a CSV file.")
    parser.add_argument('--query', nargs="+",required=True, help="The SQL query to execute.")
    parser.add_argument('--output', default='results.csv', help="The name of the output CSV file (default: results.csv).")
    parser.add_argument('--cert', default='pandb_root.crt', help="The name of the pandb ssl cert.")
    parser.add_argument('--connargs', default='pandb.json', help="The name of connection args json file.")
    parser.add_argument("--pipeline_config_file", type=Path, help="Pipeline Config File")
    return parser

if __name__ == "__main__":

    parser=parse_params()
    args, unknown_args = parser.parse_known_args()

    query = args.query
    query = " ".join([x.replace("\\","") for x in query])
    output = args.output

    cert = args.cert
    connargs = os.path.abspath(args.connargs)
    pipeline_config_file = None
    if args.pipeline_config_file:
        if Path(args.pipeline_config_file).exists():
            pipeline_config_file = str(args.pipeline_config_file)

    with open(connargs,"r") as infile:
        panargs = json.load(infile)

    encoded_password = quote_plus(panargs['password'])

    labels_dict={}
    if pipeline_config_file:
        panpipeconfig_file=str(pipeline_config_file)
        if os.path.exists(pipeline_config_file):
           print(f"{pipeline_config_file} exists.")
           with open(pipeline_config_file,'r') as infile:
               labels_dict = json.load(infile)

    if labels_dict:
        cwd = getParams(labels_dict,"CWD")
    else:
        cwd = os.path.dirname(tempfile.mkstemp()[1])

    if not os.path.dirname(output):
        output = os.path.join(cwd,output)


    # Create the connection URL
    db_url = (
        f"postgresql+psycopg2://{panargs['user']}:{encoded_password}@"
        f"{panargs['host']}:{panargs['port']}/{panargs['dbname']}?sslmode={panargs['sslmode']}&"
        f"sslrootcert={cert}"
    )

    # Connect to the database
    engine = create_engine(db_url)

    # Execute the query and retrieve data into a Pandas DataFrame
    try:
        with engine.connect() as connection:
            result = pd.read_sql_query(text(query), connection)
    except Exception as e:
        result = pd.DataFrame()

    # Save the DataFrame to a CSV file
    if not result.empty:
        result.to_csv(output, index=False)
        print(f"Data has been successfully exported to {output}.")
        output_metadata = create_metadata(output,None, metadata = {"Script":"pandb_download.py","Description":f"{query}"})
    else:
        output=None
        output_metadata=None
        print(f"Problem obtaining results for query {query}")

    if labels_dict:
        labels_dict["METADATA_FILE"]=output_metadata 
        labels_dict["OUTPUT_FILE"]=output
        export_labels(labels_dict,pipeline_config_file)



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

IS_PRESENT="^^^"
IGNORE="###"

def xcpd_proc(labels_dict,input_dir):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    command_base, container = getContainer(labels_dict,nodename="xcpd", SPECIFIC="XCPD_CONTAINER",LOGGER=IFLOGGER)

    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["APPTAINERENV_TEMPLATEFLOW_HOME"]=translate_binding(command_base,TEMPLATEFLOW_HOME)
    
    IFLOGGER.info("Checking the xcpd version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    input_dir =substitute_labels(input_dir,labels_dict)
    output_dir =substitute_labels("<CWD>/xcp_out", labels_dict)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir,exist_ok=True)

    work_dir = substitute_labels("<CWD>/xcpd_work", labels_dict)
    if not os.path.exists(work_dir):
        os.makedirs(work_dir,exist_ok=True)

    # set up dwi to process just the specific dwi session
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')

    xcpd_dict={}
    xcpd_dict = updateParams(xcpd_dict,"--participant_label","<PARTICIPANT_LABEL>")
    xcpd_dict = updateParams(xcpd_dict,"--smoothing","0")
    xcpd_dict = updateParams(xcpd_dict,"--fd-thresh","0.5")
    xcpd_dict = updateParams(xcpd_dict,"--datasets","aal=/custom_atlases/aal yeobuckner=/custom_atlases/yeobuckner schaefer=/custom_atlases/schaefer")
    xcpd_dict = updateParams(xcpd_dict,"--atlases","AAL3 yeobuckner131 yeobuckner58 schaefer1000 schaefer1000b")
    
    xcpd_dict = updateParams(xcpd_dict,"--mem_gb","<BIDSAPP_MEMORY_GB>")
    xcpd_dict = updateParams(xcpd_dict,"--nthreads","<BIDSAPP_THREADS>")
    xcpd_dict = updateParams(xcpd_dict,"--omp_nthreads","1")
    xcpd_dict = updateParams(xcpd_dict,"--fs-license-file","<FSLICENSE>")
    xcpd_dict = updateParams(xcpd_dict,"-w",work_dir)

    xcpd_dict = updateParams(xcpd_dict,"--input-type", "fmriprep")
    xcpd_dict = updateParams(xcpd_dict,"--mode", "none")
    xcpd_dict = updateParams(xcpd_dict,"--abcc-qc", "n")
    xcpd_dict = updateParams(xcpd_dict,"--combine-runs", "n")
    xcpd_dict = updateParams(xcpd_dict,"--nuisance-regressors", "36P")
    xcpd_dict = updateParams(xcpd_dict,"--despike", "y")
    xcpd_dict = updateParams(xcpd_dict,"--file-format", "nifti")
    xcpd_dict = updateParams(xcpd_dict,"--linc-qc", "n")
    xcpd_dict = updateParams(xcpd_dict,"--min-coverage", "0")
    xcpd_dict = updateParams(xcpd_dict,"--motion-filter-type", "none")
    xcpd_dict = updateParams(xcpd_dict,"--output-type", "interpolated") 
    xcpd_dict = updateParams(xcpd_dict,"--warp-surfaces-native2std", "n") 
    xcpd_dict = updateParams(xcpd_dict,"--create-matrices", "all")

    # Additional params
    XCPD_OVERRIDE_PARAMS = getParams(labels_dict,"XCPD_OVERRIDE_PARAMS")
    if XCPD_OVERRIDE_PARAMS and isinstance(XCPD_OVERRIDE_PARAMS,dict):
        add_labels(XCPD_OVERRIDE_PARAMS,xcpd_dict)        

    params = ""
    for xcpd_tag, xcpd_value in xcpd_dict.items():
        if "--" in xcpd_tag and "---" not in xcpd_tag:
            if xcpd_value == IS_PRESENT:
                params=params + " " + xcpd_tag
            elif xcpd_value == IGNORE:
                IFLOGGER.info(f"Parameter {xcpd_tag} is being skipped. This has been explicitly required in configuration.")
            else:
                # we dont need = sign for xcpd just for basi;
                params = params + " " + xcpd_tag + " " + xcpd_value

        elif "-" in xcpd_tag and "--" not in xcpd_tag:
            params = params + " " + xcpd_tag + " " + xcpd_value

        else:
            print(f"xcpd tag {xcpd_tag} not valid.") 

    command=f"{command_base}"\
            " " + input_dir +\
            " " + output_dir+\
            " participant"\
            " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    denoised_bold = getGlob(os.path.join(output_dir,'sub-{}'.format(participant_label),'ses-*','func','*desc-denoised_bold.nii.gz'))
    outliers = getGlob(os.path.join(output_dir,'sub-{}'.format(participant_label),'ses-*','func','*_outliers.tsv'))
    design = getGlob(os.path.join(output_dir,'sub-{}'.format(participant_label),'ses-*','func','*design.tsv'))
    motion_confounds = getGlob(os.path.join(output_dir,'sub-{}'.format(participant_label),'ses-*','func','*motion.tsv'))
    root_output_dir = cwd

    
    out_files=[]
    out_files.insert(0,denoised_bold)
    out_files.insert(1,outliers)
    out_files.insert(2,design)
    out_files.insert(3,motion_confounds)


    return {
        "denoised_bold": denoised_bold,
        "outliers": outliers,
        "design": design,
        "motion_confounds":motion_confounds,
        "output_dir":root_output_dir,
        "out_files":out_files
    }



class xcpdInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_dir = traits.String("",desc="FMRIPREP Directory", usedefault=True)

class xcpdOutputSpec(TraitedSpec):
    denoised_bold= File(desc='denoised_bold')
    outliers = File(desc='outliers')
    design = File(desc='design')
    motion_confounds= File(desc='motion_confounds')
    output_dir = traits.String(desc="XCPD output directory")
    out_files = traits.List(desc='list of files')
    
class xcpd_pan(BaseInterface):
    input_spec = xcpdInputSpec
    output_spec = xcpdOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = xcpd_proc(
            self.inputs.labels_dict,
            self.inputs.input_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="xcpd_node",input_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(xcpd_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}") 
           
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if input_dir is None or input_dir == "":
        input_dir = substitute_labels("<FMRIPREP_OUTPUT_DIR>", labels_dict)
        
    pan_node.inputs.input_dir =  input_dir

    return pan_node





import amico
import os
import nibabel
import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path
from functools import partial
import logging
import sys

loglevel=logging.INFO
LOGGER = logging.getLogger("amico")
LOGGER.setLevel(loglevel)
formatter = logging.Formatter('%(name)s | %(asctime)s | %(levelname)s | %(message)s')
stdout_handler = logging.StreamHandler(sys.stdout)
stdout_handler.setLevel(loglevel)
stdout_handler.setFormatter(formatter)
LOGGER.addHandler(stdout_handler)

def run_amico(participant, session, qsiprep_dir, output_path=None,path_suffix="",amico_model=None,model_config=None,amico_config = None, labels_dict={}):
    qsiprep_in=os.path.join(qsiprep_dir,f"sub-{participant}")
    bval = os.path.join(qsiprep_in,f"ses-{session}","dwi",f"sub-{participant}_ses-{session}_space-T1w_desc-preproc_dwi.bval")
    bvec = os.path.join(qsiprep_in,f"ses-{session}","dwi",f"sub-{participant}_ses-{session}_space-T1w_desc-preproc_dwi.bvec")
    dwi = os.path.join(qsiprep_in,f"ses-{session}","dwi",f"sub-{participant}_ses-{session}_space-T1w_desc-preproc_dwi.nii.gz")
    mask = os.path.join(qsiprep_in,f"ses-{session}","dwi",f"sub-{participant}_ses-{session}_space-T1w_desc-brain_mask.nii.gz")

    amico.setup()
    ae = amico.Evaluation(output_path=output_path)

    if amico_config:
        with open(amico_config,"r") as infile:
            amico_config_json = json.load(infile)
        if amico_config_json:
            for itemkey,itemvalue in amico_config_json:
                ae.set_config(itemkey,itemvalue)

    ae.set_model(amico_model)
    if amico_model == 'NODDI':
        LOGGER.info(f"Running Amico Model:\n{amico_model}")
        scheme = amico.util.fsl2scheme(bval, bvec)

        ae.load_data(dwi, scheme, mask, b0_thr=2)

        if model_config:
            with open(model_config,"r") as infile:
                model_config_json = json.load(infile)

            if model_config_json:
                if "dPar" in model_config_json.keys():
                    ae.model.dPar= model_config_json["dPar"]
                if "dIso" in model_config_json.keys():
                    ae.model.dIso=model_config_json["dIso"]
                if "IC_VFs" in model_config_json.keys():
                    ae.model.dIso=model_config_json["IC_VFs"]
                if "IC_ODs" in model_config_json.keys():
                    ae.model.dIso=model_config_json["IC_ODs"]
                if "isExvivo" in model_config_json.keys():
                    ae.model.dIso=model_config_json["isExvivo"]

        model_params = ae.model.get_params()
        LOGGER.info(f"model parameters:\n{model_params}")
        ae.generate_kernels(regenerate=True)
        ae.load_kernels()
        ae.fit()
        ae.save_results(path_suffix=path_suffix)

def parse_params():
    parser = ArgumentParser(description="amico")
    parser.add_argument("--participant", help="BIDS Participant label")
    parser.add_argument("--session", help="BIDS Session label")
    parser.add_argument("--qsiprep_dir", type=Path,  help="Input qsiprep directory")
    parser.add_argument("--amico_model", help="Amico model to run", default="NODDI")
    parser.add_argument("--path_suffix", help="path suffix")
    parser.add_argument("--model_config", type=Path, help="Model config json")
    parser.add_argument("--amico_config", type=Path, help="AMICO configuration json")
    parser.add_argument("--output_path", type=Path, help="Output results")
    parser.add_argument("--pipeline_config_file", type=Path, help="Pipeline Config File")
    return parser

def main():
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()
    participant = args.participant
    session = args.session
    amico_model = args.amico_model
    path_suffix = args.path_suffix
    qsiprep_dir = args.qsiprep_dir
    model_config = args.model_config
    amico_config = args.amico_config
    output_path = args.output_path

    if not output_path:
        output_path=os.path.join(os.getcwd,"amico_out")

    if not os.path.exists(output_path):
        os.makedirs(output_path,exist_ok=True)

    pipeline_config_file = None
    if args.pipeline_config_file:
        if Path(args.pipeline_config_file).exists():
            pipeline_config_file = str(args.pipeline_config_file)

    labels_dict={}
    if pipeline_config_file:
        panpipeconfig_file=str(pipeline_config_file)
        if os.path.exists(pipeline_config_file):
           print(f"{pipeline_config_file} exists.")
           with open(pipeline_config_file,'r') as infile:
               labels_dict = json.load(infile)

    run_amico(participant, session, qsiprep_dir, output_path=output_path, amico_model=amico_model, path_suffix=path_suffix, model_config=model_config,amico_config = amico_config,labels_dict=labels_dict)

# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()


from panpipelines.utils.util_functions import *
from panpipelines.pipelines.panpipeline import *
import os
import glob
import panpipelines.workflows.collatecsvgroup_workflow as panworkflow

TARGET_ANALYSIS_LEVEL=["group"]

class collatecsvgroup_panpipeline(panpipeline):

    def __init__(self,labels_dict,pipeline_dir, participant_label, name='collatecsvgroup_panpipeline',createGraph=True,LOGGER=None,execution={},analysis_level = "group", participant_project=None,participant_session=None):

        # perhaps check here that the analysis_level is indeed group!
        if analysis_level not in TARGET_ANALYSIS_LEVEL:
            if LOGGER:
                LOGGER.error(f"pipeline {name} not defined for {analysis_level}")
                raise ValueError(f"pipeline {name} not defined for {analysis_level}")

        super().__init__(labels_dict,pipeline_dir, participant_label,name,createGraph,LOGGER,execution,analysis_level, participant_project,participant_session,panworkflow)








import pandas as pd
import numpy as np
import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path
from functools import partial
import getpass
import xnat
from pydicom import dcmread
from  pydicom.multival import MultiValue as MultiValue
import datetime
import os
from collections import OrderedDict
import fnmatch
from panpipelines.utils.util_functions import *

INCLUDE_QC=False

PAN_SCANTYPES = {
    "DWI":["DTI_MB_3SHELL_mono_TE91_ORIG","DTI_MB_3SHELL 0 1000 2000 mono te 91_ORIG","DTI_MB_3SHELL_0_1K_2K"],
    "DWI_RPE" : ["DTI_MB_3SHELL RPE_ORIG","DTI_MB_3SHELL_RPE"],
    "MPRAGE" : ["Accelerated Sagittal MPRAGE","Sagittal 3D Accelerated MPRAGE","Accelerated Sagittal MPRAGE (MSV21)"],
    "RSFMRI" : ["RS-FMRI-MB-high_resolution","Axial fcMRI _EYES OPEN","Axial fcMRI _EYES_Open_plus"],
    "FLAIR" : ["Sagittal 3D FLAIR","Sagittal 3D FLAIR","Sagittal 3D FLAIR (MSV21)"] ,
    "PCASL" : ["WIP SOURCE - Axial 3D PCASL"],
    "PCASL_M0" : ["Axial 3D M0"],
    "PASL" : ["PASL_PRODUCT_PLD-2750"],
    "PASL_M0" : ["PASL_PRODUCT_PLD-2750 MO"],
    "PASL_LEGACY": ["AXIAL 3D PASL plus M0","Axial 3D PASL plus M0"],
    "MEGRE" : ["Axial 3D ME T2 GRE","Axial 3D ME T2 GRE_noT2*"],
    "T2HIPPO" : ["HighResHippocampus","HighResHippoObliqueperpenHIPPO"],
    "T2SPACE" : ["Sagittal 3D T2 SPACE","3D_Brain_VIEW_T2"]
}

PAN_OPTIONAL = ["T2HIPPO"]
PAN_LEGACY = ["T2SPACE","PASL_LEGACY"]
PASL_OPTIONAL = ["PCASL","PCASL_M0"]
PCASL_OPTIONAL = ["PASL","PASL_M0"]

CANDO_STRUCT_REQS = ["MPRAGE"]

CANDO_PASL_REQS = ["PASL","PASL_M0","MPRAGE"]
CANDO_PASLSDC_REQS = ["PASL","PASL_M0","MPRAGE","MEGRE"]
CANDO_PASLSDC_FMRI_REQS = ["PASL","PASL_M0","MPRAGE","MEGRE","RSFMRI"]
CANDO_PASLSDC_DWI_REQS = ["PASL","PASL_M0","MPRAGE","DWI","DWI_RPE"]

CANDO_PASLLEGACY_REQS = ["PASL_LEGACY","MPRAGE"]
CANDO_PASLLEGACYSDC_REQS = ["PASL_LEGACY","MPRAGE","MEGRE"]
CANDO_PASLLEGACYSDC_FMRI_REQS = ["PASL_LEGACY","MPRAGE","MEGRE","RSFMRI"]
CANDO_PASLLEGACYSDC_DWI_REQS = ["PASL_LEGACY","MPRAGE","DWI","DWI_RPE"]

CANDO_PCASL_REQS = ["PCASL","PCASL_M0","MPRAGE"]
CANDO_PCASLSDC_REQS = ["PCASL","PCASL_M0","MPRAGE","MEGRE"]
CANDO_PCASLSDC_FMRI_REQS = ["PCASL","PCASL_M0","MPRAGE","MEGRE","RSFMRI"]
CANDO_PCASLSDC_DWI_REQS = ["PCASL","PCASL_M0","MPRAGE","DWI","DWI_RPE"]

CANDO_DWI_REQS = ["DWI","MPRAGE"]
CANDO_DWISDC_REQS = ["DWI","DWI_RPE","MPRAGE"]

CANDO_RSFMRI_REQS = ["RSFMRI","MPRAGE"]
CANDO_RSFMRISDC_REQS = ["RSFMRI","MPRAGE","MEGRE"]
CANDO_RSFMRISDC_DWI_REQS = ["RSFMRI","MPRAGE","DWI","DWI_RPE"]

TARGET_PROJECT="PAN_250_1"

###########################################################



PAN_RELEASE="pan_release"
PAN_SITE="pan_site"
SUBJECT_LABEL="xnat_subject_label"
HMLID="hml_id"
XNATID_ORIG="orig_xnat_subject_label"
MCID="participant_id_parent"
P2ID = "p2_id"

ORIG_PROJECT="orig_pan_site"
GENDER = "gender"
YOB = "yob"
SESSION_LABEL = "session_label"
SESSION_MODALITY = "session_modality"
SESSION_SCANDATE="session_scandate"
SESSION_NOTES="session_notes"
BIDS_SUBJECT="subject_id"
BIDS_SESSION="session_id"
CANDO_STRUCT = "cando_struct"
CANDO_DWI = "cando_dwi"
CANDO_DWI_SDC = "cando_dwi_sdc"
CANDO_PASL = "cando_pasl"
CANDO_PASL_SDC = "cando_pasl_sdc"
CANDO_PASL_SDC_FMRI= "cando_pasl_sdc_fmri"
CANDO_PASL_SDC_DWI= "cando_pasl_sdc_dwi"
CANDO_PASLLEGACY = "cando_pasllegacy"
CANDO_PASLLEGACY_SDC = "cando_pasllegacy_sdc"
CANDO_PASLLEGACY_SDC_FMRI= "cando_pasllegacy_sdc_fmri"
CANDO_PASLLEGACY_SDC_DWI= "cando_pasllegacy_sdc_dwi"
CANDO_PCASL = "cando_pcasl"
CANDO_PCASL_SDC = "cando_pcasl_sdc"
CANDO_PCASL_SDC_FMRI = "cando_pcasl_sdc_fmri"
CANDO_PCASL_SDC_DWI = "cando_pcasl_sdc_dwi"
CANDO_RSFMRI = "cando_rsfmri"
CANDO_RSFMRI_SDC = "cando_rsfmri_sdc"
CANDO_RSFMRI_SDC_DWI = "cando_rsfmri_sdc_dwi"
MRI_SCAN_ORDER = "mri_scan_order"
MRI_MISSING_SCANS = "mri_missing_scans"
MRI_MISSING_OPTIONAL_SCANS = "mri_missing_optional_scans"
MRI_PROBLEM_SCANS = "mri_problem_scans"
BIDS_AVAIL = "bids_avail"
BIDSQC_STATUS = "bidsqc_status"
BIDSQC_FAIL = "bidsqc_fail"
MRIQC_AVAIL = "mriqc_avail"
QSIPREP_AVAIL = "qsiprep_avail"
EDDYQC_AVAIL = "eddyqc_avail"
ASLPREP_AVAIL = "aslprep_avail"
FMRIPREP_AVAIL = "fmriprep_avail"
RADNORMAL = "radnormal"

CANDO_ASL = "cando_asl"
CANDO_ASL_SDC = "cando_asl_sdc"
CANDO_ASL_SDC_FMRI= "cando_asl_sdc_fmri"
CANDO_ASL_SDC_DWI= "cando_asl_sdc_dwi"

MANUFACTURER = "manufacturer"
STATION_NAME="station_name"
DEVICE_SERIAL_NUMBER="device_serial_number"
SOFTWARE_VERSION="software_versions"
MANUFACTURER_MODEL_NAME="manufacturer_model_name"
IMPLEMENTATION_VERSION_NAME="implementation_version_name"

CREATION_DATE="row_creation_datetime"


table_header = []
#table_header.append(PAN_RELEASE)
table_header.append(PAN_SITE)
table_header.append(SUBJECT_LABEL)
table_header.append(HMLID)
#table_header.append(XNATID_ORIG)
#table_header.append(MCID)
#table_header.append(P2ID)
#table_header.append(ORIG_PROJECT)
#table_header.append(GENDER)
#table_header.append(YOB)
table_header.append(SESSION_LABEL)
table_header.append(SESSION_MODALITY)
table_header.append(SESSION_SCANDATE)
table_header.append(SESSION_NOTES)
table_header.append(BIDS_SUBJECT)
table_header.append(BIDS_SESSION)
table_header.append(CANDO_STRUCT)
table_header.append(CANDO_DWI)
table_header.append(CANDO_DWI_SDC)
table_header.append(CANDO_PASL)
table_header.append(CANDO_PASL_SDC)
table_header.append(CANDO_PASL_SDC_FMRI)
table_header.append(CANDO_PASL_SDC_DWI)
table_header.append(CANDO_PASLLEGACY)
table_header.append(CANDO_PASLLEGACY_SDC)
table_header.append(CANDO_PASLLEGACY_SDC_FMRI)
table_header.append(CANDO_PASLLEGACY_SDC_DWI)
table_header.append(CANDO_PCASL)
table_header.append(CANDO_PCASL_SDC)
table_header.append(CANDO_PCASL_SDC_FMRI)
table_header.append(CANDO_PCASL_SDC_DWI)
table_header.append(CANDO_RSFMRI)
table_header.append(CANDO_RSFMRI_SDC)
table_header.append(CANDO_RSFMRI_SDC_DWI)
table_header.append(MRI_SCAN_ORDER)
table_header.append(MRI_MISSING_SCANS)
table_header.append(MRI_MISSING_OPTIONAL_SCANS)
table_header.append(MRI_PROBLEM_SCANS)
table_header.append(BIDSQC_STATUS)
table_header.append(BIDSQC_FAIL)
table_header.append(BIDS_AVAIL)
table_header.append(MRIQC_AVAIL)
table_header.append(QSIPREP_AVAIL)
table_header.append(EDDYQC_AVAIL)
table_header.append(ASLPREP_AVAIL)
table_header.append(FMRIPREP_AVAIL)
table_header.append(MANUFACTURER)
table_header.append(STATION_NAME)
table_header.append(DEVICE_SERIAL_NUMBER)
table_header.append(SOFTWARE_VERSION)
table_header.append(MANUFACTURER_MODEL_NAME)
table_header.append(IMPLEMENTATION_VERSION_NAME)
table_header.append(CREATION_DATE)

if INCLUDE_QC:
    mriqc_headers=[]
    mriqc_headers_t1w=["cjv","cnr","efc","fber","fwhm_avg","qi_1","qi_2","snr_total"]
    mriqc_t1w_label = "rec-defaced_T1w"
    mriqc_headers_t2w=["cjv","cnr","efc","fber","fwhm_avg","qi_1","qi_2","snr_total"]
    mriqc_t2w_label = "acq-hippo_T2w"
    mriqc_headers_bold=["efc","fber","fd_mean","gcor","gsr_x","gsr_y","snr","tsnr"]
    mriqc_bold_label = "task-rest_bold"            
    mriqc_headers.extend([f"MRIQC.{mriqc_t1w_label}.{x}" for x in mriqc_headers_t1w])
    mriqc_headers.extend([f"MRIQC.{mriqc_t2w_label}.{x}" for x in mriqc_headers_t2w])
    mriqc_headers.extend([f"MRIQC.{mriqc_bold_label}.{x}" for x in mriqc_headers_bold])
    table_header.extend(mriqc_headers)

    qsiprepqc_headers=[]
    qsiprepqc_headers_imageQC=["mean_fd","max_fd","max_rotation","max_translation","raw_num_bad_slices","t1_num_bad_slices"]
    qsiprepqc_imageQC_label="desc-ImageQC_dwi"
    qsiprepqc_headers.extend([f"QSIPREPQC.{qsiprepqc_imageQC_label}.{x}" for x in qsiprepqc_headers_imageQC])
    table_header.extend(qsiprepqc_headers)

    eddyqc_headers=[]
    eddyqc_headers_qc=["qc_mot_abs","qc_mot_rel","qc_outliers_tot","qc_cnr_avg"]
    eddyqc_qc_label="qc"            
    eddyqc_headers.extend([f"EDDYQC.{eddyqc_qc_label}.{x}" for x in eddyqc_headers_qc])
    table_header.extend(eddyqc_headers)

    aslprepqc_headers=[]
    aslprepqc_headers_cbf=["FD", "cbfQEI","GMmeanCBF","WMmeanCBF","Gm_Wm_CBF_ratio","NEG_CBF_PERC"]
    aslprepqc_cbf_label="desc-qualitycontrol_cbf"           
    aslprepqc_headers.extend([f"ASLPREPQC.{aslprepqc_cbf_label}.{x}" for x in aslprepqc_headers_cbf])
    table_header.extend(aslprepqc_headers)

table_header.append(RADNORMAL)


##################################################################

def getProject(p2site):
    if p2site:
        if p2site == "UA":
            return "001_HML"
        elif p2site == "UM":
            return "002_HML"
        elif p2site == "EU":
            return "003_HML"
        elif p2site == "JH":
            return "004_HML"
        elif p2site == "PAN250":
            return "PAN_250_1"
        else:
            print(f"P2site {p2site} not recognized")
            return ''
    else:
        print("P2site not defined.")
        return ''

def getSite(project):
    if project:
        if project == "001_HML":
            return "UA"
        elif project == "002_HML":
            return "UM"
        elif project == "003_HML":
            return "EU"
        elif project == "004_HML":
            return "JH"
        elif project == "PAN_250_1":
            return "PAN250"
        else:
            print(f"Project {project} not recognized")
            return ''
    else:
        print("P2site not defined.")
        return ''


def getParams(pardict, key, default=None):
    if key is not None and pardict is not None:
        if key in pardict:
            if not pardict[key] and default:
                return default
            else:
                return pardict[key]
    return default

def loadParams(pardict, key, value, update=True):
    if key and value:
        if key in pardict:
            if not pardict[key] or update:
                pardict[key]=value
    #    else:
    #        pardict[key]=value
    return pardict

def cleanServer(server):
    server.strip()
    if server[-1] == '/':
        server = server[:-1]
    if server.find('http') == -1:
        server = 'https://' + server
    return server

def _path_exists(path, parser):
    """Ensure a given path exists."""
    if path is None or not Path(path).exists():
        raise parser.error(f"Path does not exist: <{path}>.")
    return Path(path).expanduser().absolute()

def parse_params():
    parser = ArgumentParser(description="BIDSQC")
    PathExists = partial(_path_exists, parser=parser)
    parser.add_argument("csvout", type=Path, help="The directory where output files are stored")
    parser.add_argument("--projects", help="Projects", required=False, nargs='*')
    parser.add_argument("--excluded_participants", help="Participants to exclude", required=False, nargs='*')
    parser.add_argument("--host", help="XNAT host", required=True)
    parser.add_argument("--credentials", type=PathExists, help="credential file")
    parser.add_argument("--pipeline_config_file", type=Path, help="Pipeline Config File")
    return parser


def getMissingScans(sequence,project=None):
    if project:
        if project == "004_HML":
            return list(set(list(PAN_SCANTYPES.keys())).difference(sequence).difference(PAN_OPTIONAL).difference(PCASL_OPTIONAL).difference(PAN_LEGACY))
        else:
            return list(set(list(PAN_SCANTYPES.keys())).difference(sequence).difference(PAN_OPTIONAL).difference(PASL_OPTIONAL).difference(PAN_LEGACY))
    else: 
        return list(set(list(PAN_SCANTYPES.keys())).difference(sequence).difference(PAN_OPTIONAL).difference(PAN_LEGACY))

def getMissingOptionalScans(sequence,project=None):
    if project:
        if project == "004_HML":
            return list(set(list(PAN_SCANTYPES.keys())).difference(sequence).difference(PCASL_OPTIONAL).difference(PAN_LEGACY).intersection(PAN_OPTIONAL))
        else:
            return list(set(list(PAN_SCANTYPES.keys())).difference(sequence).difference(PASL_OPTIONAL).difference(PAN_LEGACY).intersection(PAN_OPTIONAL))
    else: 
        return list(set(list(PAN_SCANTYPES.keys())).difference(sequence).difference(PAN_LEGACY).intersection(PAN_OPTIONAL))

def findPANScanType(scantype):
    for pan_type, pan_val in PAN_SCANTYPES.items():
        if scantype in pan_val:
            return pan_type
    return None

def getSequence(experiment):

    # roundabout way to do this as iterator on scans in xnatpy is a little wonky
    scans = experiment.scans
    sequence=[]
    for scan_index in range(len(scans)):
        scan = experiment.scans[scan_index]
        scan_id = scan.id
        scan_type = scan.type
        pan_type = findPANScanType(scan_type)
        if pan_type:
            sequence.append(pan_type)

    # hack for interoperability scans
    if len(sequence) > 50:
        sequence = list(set(sequence))

    return sequence


def cando_proc(REQS,sequence,missingscans, problemscans=[],struct_sessions={}):
    validscans = list(set(sequence).difference(missingscans).difference(problemscans))
    missing = list(set(REQS).difference(validscans))
    if not missing:
        return "Y"
    elif len(missing) == 1 and missing[0] == "MPRAGE" and struct_sessions and structsAvailable(struct_sessions):
        return "S"
    else:
        return "N"

def structsAvailable(struct_sessions):
    structAvail=False
    for itemkey, itemvalue in struct_sessions.items():
        if itemvalue == "Y":
            structAvail = True
    return structAvail


def structsOverSessions(experiments,orig_project):
    struct_sessions={}

    for exp_index in range(len(experiments)):
        experiment = experiments[exp_index]
        exp_label = experiment.label
        if experiment.modality == 'MR' or '_MR_' in experiment.label:
            sequence = getSequence(experiment)
            missing_scans = getMissingScans(sequence,orig_project)
            struct_sessions[exp_label] = cando_proc(CANDO_STRUCT_REQS,sequence,missing_scans)

    return struct_sessions


def getQCMeasures(files, headers, table_row):

    json_dict = {}
    for header in headers:
        sourcename = header.split(".")[1]
        if sourcename not in json_dict.keys():
            json_dict[sourcename] = None
            for file_index in range(len(files)):
                file = files[file_index]
                if f"{sourcename}.json" in file.name:
                    file_name=os.path.join("/tmp",os.path.basename(file.name))
                    file.download(file_name)
                    with open(file_name,'r') as in_file:
                        file_json = json.load(in_file)
                    if file_json:
                        json_dict[sourcename] = file_json
                    else:
                        json_dict[sourcename] = None
                    break


    for header in headers:
        prefix = header.split(".")[0]
        sourcename = header.split(".")[1]
        field = header.split(".")[2]
        file_json = json_dict[sourcename]
        if file_json is not None:
            if field in file_json.keys():
                loadParams(table_row,header,str(file_json[field]))
            else:
                loadParams(table_row,header,"")
        else:
            loadParams(table_row,header,"")

    return table_row


def getQCMeasuresCSV(files, headers, table_row):


    json_dict = {}
    for header in headers:
        sourcename = header.split(".")[1]
        if sourcename not in json_dict.keys():
            json_dict[sourcename] = None
            for file_index in range(len(files)):
                file = files[file_index]
                if f"{sourcename}.csv" in file.name:
                    file_name=os.path.join("/tmp",os.path.basename(file.name))
                    file.download(file_name)
                    if os.path.exists(file_name):
                        df = pd.read_table(file_name,sep=",")
                        json_dict[sourcename] = df
                    break


    for header in headers:
        prefix = header.split(".")[0]
        sourcename = header.split(".")[1]
        field = header.split(".")[2]
        data_f = json_dict[sourcename]
        if data_f is not None:
            if field in data_f.columns.tolist():
                measure_f = df[field].iloc[0]
                loadParams(table_row,header,str(measure_f))
            else:
                loadParams(table_row,header,"")
        else:
            loadParams(table_row,header,"")

    return table_row


def appendText(comments,message,seperator=","):
    if comments:
        comments = comments + ": " + message
    else:
        comments = message
    return comments

def loadText(comments,message,replace = True):
    if not comments:
        comments=message
    elif replace:
        comments=message
    return comments

def appendTableRows(table_rows,table_data):    
    for table_row in table_rows:
        new_table_row = []
        for itemkey, itemvalue in table_row.items():
            new_table_row.append(str(itemvalue))
        table_data.append(new_table_row)
    return table_data

def initializeRow():
    tablerow = OrderedDict()
    for column in table_header:
        tablerow[column]=""
    return tablerow

def getSubjectRows(xnatid, connection, table_header,project=None, table_row={},all_dups=[], hmlid="",do_share=False):
    table_rows = []
    return table_rows

def iterateSessions(connection,xnatid,table_row={},all_dups=[],hmlid="",do_share=True):
    table_rows = []
    return table_rows


def getSubjectCustomField(connection, subject, field):
    try:
        return  subject.fields[field]
    except:
        print(f"problem obtaining subject custom field : {field}")
        if field == "hmlid":
            return subject.label
        else:
            return ""

def canProc(df):
    dfy = df == "Y"
    dfs = df == "S"
    if pd.DataFrame.any(dfy):
        return "Y"
    elif pd.DataFrame.any(dfs):
        return "S"
    else:
        return "N"


def canProcList(dflist):
    df = pd.concat(dflist,axis=1)
    dfy = df == "Y"
    dfs = df == "S"
    if pd.DataFrame.any(pd.DataFrame.any(dfy)):
        return "Y"
    elif pd.DataFrame.any(pd.DataFrame.any(dfs)):
        return "S"
    else:
        return "N"


def getBidsQC(host,user,password,projects,csvout,excluded_participants=[],LOGFILE=None,pipeline_config_file=None):
  
    labels_dict={}
    if pipeline_config_file:
        panpipeconfig_file=str(pipeline_config_file)
        if os.path.exists(pipeline_config_file):
           print(f"{pipeline_config_file} exists.")
           with open(pipeline_config_file,'r') as infile:
               labels_dict = json.load(infile)


    bidsqcOutdir=os.path.abspath(os.path.dirname(csvout))

    if not os.path.isdir(bidsqcOutdir):
        os.makedirs(bidsqcOutdir,exist_ok=True)

    try:
        table_data = []

        scantest = os.path.join("/tmp",'scantest.dcm')

        with xnat.connect(server=host,user=user,password=password) as connection:

            if not projects:
                projects = connection.projects
            
            if len(projects) == 1:
                pan_release_t = projects[0]
            else:
                pan_release_t = "Latest"

            for PROJ in list(set(projects)):
                project = connection.projects[PROJ]

                project_t = project.id

                bidsqcOutProjdir=os.path.join(bidsqcOutdir,project.id)
                if not os.path.isdir(bidsqcOutProjdir):
                    os.mkdir(bidsqcOutProjdir)

                bidsqcDetailsOutdir=os.path.join(bidsqcOutProjdir,'Details')
                if not os.path.isdir(bidsqcDetailsOutdir):
                    os.mkdir(bidsqcDetailsOutdir)

                subjects = project.subjects
                for sub_index in range(len(subjects)):

                    subject_row = []

                    table_row = initializeRow()
                    loadParams(table_row,PAN_RELEASE,pan_release_t)
                    loadParams(table_row,PAN_SITE,project_t)

                    subject = subjects[sub_index]

                    subject_t = subject.label
                    loadParams(table_row,SUBJECT_LABEL,subject_t)

                    gender_t = subject.demographics.gender
                    loadParams(table_row,GENDER,gender_t)

                    loadParams(table_row,HMLID,getSubjectCustomField(connection,subject,"hmlid"))
                    loadParams(table_row,XNATID_ORIG,getSubjectCustomField(connection,subject,"xnatorigid"))
                    loadParams(table_row,MCID,getSubjectCustomField(connection,subject,"mcid"))
                    loadParams(table_row,P2ID,getSubjectCustomField(connection,subject,"p2id"))
                    orig_project = getSubjectCustomField(connection,subject,"site")
                    loadParams(table_row,ORIG_PROJECT,orig_project)

                    yob_t = ""
                    dob = subject.demographics.dob
                    if dob:
                        yob_t = datetime.datetime.strftime(dob,"%Y")
                    loadParams(table_row,YOB,yob_t)

                    experiment_t = 'None'
                    loadParams(table_row,SESSION_LABEL,experiment_t)

                    Manufacturer_t=""
                    StationName_t = ""
                    DeviceSerialNumber_t = ""
                    SoftwareVersions_t=""
                    ManufacturerModelName_t=""
                    ImplementationversionName_t=""

                    experiments = subject.experiments

                    if len(experiments) < 1:
                        subject_row.append(table_row)
                      
                    struct_sessions=structsOverSessions(experiments,orig_project)

                    for exp_index in range(len(experiments)):

                        session_table_row = table_row.copy()

                        experiment = experiments[exp_index]
                        experiment_t = experiment.label
                        loadParams(session_table_row,SESSION_LABEL,experiment_t)

                        
                        if experiment.modality == 'MR' or '_MR_' in experiment.label:

                            loadParams(session_table_row,SESSION_MODALITY,"MR")

                            scandate_t = datetime.datetime.strftime(experiment.date,"%Y%m%d")
                            loadParams(session_table_row,SESSION_SCANDATE,scandate_t)
                         
                            scans = experiment.scans

                            scan_sequence = getSequence(experiment)
                            loadParams(session_table_row,MRI_SCAN_ORDER,"|".join(scan_sequence))

                            missing_scans = getMissingScans(scan_sequence,orig_project)
                            loadParams(session_table_row,MRI_MISSING_SCANS,"|".join(missing_scans))

                            missing_optional_scans = getMissingOptionalScans(scan_sequence,orig_project)
                            loadParams(session_table_row,MRI_MISSING_OPTIONAL_SCANS,"|".join(missing_optional_scans))


                            for scan_index in range(len(scans)):
                                if "DICOM" in experiment.scans[scan_index].resources.keys():
                                    try:
                                        experiment.scans[scan_index].resources['DICOM'].files[0].download(scantest)
                                        ds = dcmread(scantest)
                                        if not gender_t:
                                            gender_t = ds.PatientSex
                                            loadParams(session_table_row,GENDER,gender_t)

                                        if not yob_t:
                                            if ds.PatientBirthDate:
                                                agedate=datetime.datetime.strptime(ds.PatientBirthDate,"%Y%m%d")
                                                yob_t = datetime.datetime.strftime(agedate,"%Y")
                                                loadParams(session_table_row,YOB,yob_t)

                                        if not scandate_t:
                                            dcm_scandate = ds.StudyDate
                                            scandate = datetime.datetime.strptime(dcm_scandate,"%Y%m%d")
                                            scandate_t = datetime.datetime.strftime(scandate,"%Y%m%d")
                                            loadParams(session_table_row,SESSION_SCANDATE,scandate_t)

                                        if not Manufacturer_t:
                                            Manufacturer_t=ds.Manufacturer
                                            loadParams(session_table_row,MANUFACTURER,Manufacturer_t) 
                                        if not StationName_t:
                                            StationName_t = ds.StationName
                                            loadParams(session_table_row,STATION_NAME,StationName_t) 
                                        if not DeviceSerialNumber_t:
                                            DeviceSerialNumber_t = ds.DeviceSerialNumber
                                            loadParams(session_table_row,DEVICE_SERIAL_NUMBER,DeviceSerialNumber_t) 
                                        if not SoftwareVersions_t:
                                            SoftwareVersions_t = ds.SoftwareVersions 
                                            if isinstance(SoftwareVersions_t,MultiValue):
                                                SoftwareVersions_t = list(SoftwareVersions_t)[0]
                                            loadParams(session_table_row,SOFTWARE_VERSION,SoftwareVersions_t) 
                                        if not ManufacturerModelName_t:
                                            ManufacturerModelName_t = ds.ManufacturerModelName
                                            loadParams(session_table_row,MANUFACTURER_MODEL_NAME,ManufacturerModelName_t) 
                                        if not ImplementationversionName_t:
                                            ImplementationversionName_t = ds.file_meta.ImplementationVersionName
                                            loadParams(session_table_row,IMPLEMENTATION_VERSION_NAME,ImplementationversionName_t) 
                                        break

                                    except Exception as e:
                                        pass


                            notes_t = experiment.note
                            loadParams(session_table_row,SESSION_NOTES,notes_t)


                            # get blank mriqc measures for situations where MRIQC resource doesn't exist
                            if INCLUDE_QC:
                                session_table_row = getQCMeasures([],mriqc_headers,session_table_row)
                                session_table_row = getQCMeasuresCSV([],qsiprepqc_headers,session_table_row)
                                session_table_row = getQCMeasures([],eddyqc_headers,session_table_row)
                                session_table_row = getQCMeasuresCSV([],aslprepqc_headers,session_table_row)

                            radread_t=""
                            assessors = experiment.assessors
                            for assessor_index in range(len(assessors)):
                                assessor = assessors[assessor_index]
                                if assessor.__xsi_type__ == "rad:radiologyReadData":
                                    finding = assessor.get("finding/normal_status")
                                    if not radread_t:
                                        radread_t = str(finding)
                                        loadParams(session_table_row,RADNORMAL,radread_t)
                                    else:
                                        radread_t = radread_t + " : " + str(finding)
                                        loadParams(session_table_row,RADNORMAL,radread_t)

                            resources = experiment.resources
                            problemscans=[]
                            for res_index in range(len(resources)):
                                resource = resources[res_index]

                                if resource.label == 'BIDS-AACAZ':
                                    bids_participant_id=""
                                    bids_session_id = ""
                                    files = resource.files
                                    numfiles = len(files)
                                    for fileid in range(numfiles):
                                        file=files[fileid]
                                        if fnmatch.fnmatch(file.path,'sub-*/ses-*/*'):
                                            bids_participant_id = file.path.split('/')[0]
                                            bids_session_id = file.path.split('/')[1]
                                            break

                                    if not bids_participant_id:
                                        for fileid in range(numfiles):
                                            file=files[fileid]
                                            if fnmatch.fnmatch(file.path,'sub-*/*'):
                                                bids_participant_id = file.path.split('/')[0]
                                                bids_session_id = ''
                                                break

                                    loadParams(session_table_row,BIDS_SUBJECT,bids_participant_id)
                                    loadParams(session_table_row,BIDS_SESSION,bids_session_id)


                                if resource.label == 'BIDSQC-AACAZ':
                                    fail_t = [x for x in resource.files if 'FAIL.txt' in x]
                                    if len(fail_t) > 0:
                                        bidsqc_t = 'FAIL'
                                        loadParams(session_table_row,BIDSQC_STATUS,bidsqc_t)
                                        files = resource.files
                                        for file_index in range(len(files)):
                                            file = files[file_index]
                                            if file.name == 'FAIL.txt':
                                                fail_file = os.path.join(bidsqcDetailsOutdir,'{}_{}_FAIL.txt'.format(subject_t,experiment_t))
                                                file.download(fail_file)
                                                with open(fail_file,'r') as in_file:
                                                    fail_text_lines = in_file.readlines()
                                                details_t = ''.join(fail_text_lines).replace('\n\n',' | ').replace('\n','')
                                                loadParams(session_table_row,BIDSQC_FAIL,details_t)

                                    pass_t = [x for x in resource.files if 'PASS.txt' in x]
                                    if len(pass_t) > 0:
                                        bidsqc_t = 'PASS'
                                        loadParams(session_table_row,BIDSQC_STATUS,bidsqc_t)
                                      
                                    # attempt heuristic here
                                    if bidsqc_t == "FAIL":
                                        details_parts = details_t.split("|")
                                        hippo_missing = [x for x in details_parts if "acq-hippo_T2w.json json file not found" in x]
                                        pasl_pe = [x for x in details_parts if "acq-prod_asl.json::PhaseEncodingDirection is *DIFFERENT*" in x]
                                        dwirpe_pe = [x for x in details_parts if "dir-AP_epi.json::PhaseEncodingDirection is *DIFFERENT*" in x]
                                        rsfmri_pe = [x for x in details_parts if "task-rest_bold.json::PhaseEncodingDirection is *DIFFERENT*" in x]

                                        if (len(details_parts) == 1 and hippo_missing) or (len(details_parts) == 2 and hippo_missing and not details_parts[1].strip()):
                                            bidsqc_t = 'PASS'
                                            loadParams(session_table_row,BIDSQC_STATUS,bidsqc_t)

                                        if pasl_pe:
                                            problemscans.append("PASL")

                                        if dwirpe_pe:
                                            problemscans.append("DWI_RPE")

                                        elif rsfmri_pe:
                                            problemscans.append("RSFMRI")


                                bids_t = "N"
                                if resource.label == 'BIDS-AACAZ' and resource.files is not None and len(resource.files) > 0:
                                    bids_t = "Y"
                                    loadParams(session_table_row,BIDS_AVAIL,bids_t)

                                mriqc_t = "N"
                                if resource.label == 'MRIQC-AACAZ' and resource.files is not None and len(resource.files) > 0:
                                    mriqc_t = "Y"
                                    loadParams(session_table_row,MRIQC_AVAIL,mriqc_t)
                                    if INCLUDE_QC:
                                        session_table_row = getQCMeasures(resource.files,mriqc_headers,session_table_row)

                                fmriprep_t = "N"
                                if resource.label == 'FMRIPREP-AACAZ' and resource.files is not None and len(resource.files) > 0:
                                    fmriprep_t = "Y"
                                    loadParams(session_table_row,FMRIPREP_AVAIL,fmriprep_t)

                                qsiprep_t = "N"
                                if resource.label == 'QSIPREP-AACAZ' and resource.files is not None and len(resource.files) > 0:
                                    qsiprep_t = "Y"
                                    loadParams(session_table_row,QSIPREP_AVAIL,qsiprep_t)
                                    if INCLUDE_QC:
                                        session_table_row = getQCMeasuresCSV(resource.files,qsiprepqc_headers,session_table_row)

                                eddyqc_t = "N"
                                if resource.label == 'EDDYQC-AACAZ' and resource.files is not None and len(resource.files) > 0:
                                    eddyqc_t = "Y"
                                    loadParams(session_table_row,EDDYQC_AVAIL,eddyqc_t)
                                    if INCLUDE_QC:
                                        session_table_row = getQCMeasures(resource.files,eddyqc_headers,session_table_row)

                                aslprep_t = "N"
                                if resource.label == 'ASLPREP-AACAZ' and resource.files is not None and len(resource.files) > 0:
                                    aslprep_t = "Y"
                                    loadParams(session_table_row,ASLPREP_AVAIL,aslprep_t)
                                    if INCLUDE_QC:
                                        session_table_row= getQCMeasuresCSV(resource.files,aslprepqc_headers,session_table_row)

                            # address problem scans and analysis possibility
                            loadParams(session_table_row,MRI_PROBLEM_SCANS,"|".join(problemscans))


                        elif experiment.modality == 'US' or '_US_' in experiment.label:
                            loadParams(session_table_row,SESSION_MODALITY,"US")

                            scandate_t = datetime.datetime.strftime(experiment.date,"%Y%m%d")
                            loadParams(session_table_row,SESSION_SCANDATE,scandate_t)
                        else:
                            loadParams(session_table_row,SESSION_MODALITY,"UNKNOWN")

                            scandate_t = datetime.datetime.strftime(experiment.date,"%Y%m%d")
                            loadParams(session_table_row,SESSION_SCANDATE,scandate_t)


                        if getParams(session_table_row,SESSION_MODALITY) == "MR":
                            struct_sessions[experiment_t] = cando_proc(CANDO_STRUCT_REQS,scan_sequence,missing_scans,problemscans=problemscans,struct_sessions=struct_sessions)

                        
                        subject_row.append(session_table_row)

                    for session_row in subject_row:
                        if getParams(session_row,SESSION_MODALITY) == "MR":
                            session_sequence = getParams(session_row,MRI_SCAN_ORDER).split("|")
                            session_missing = getParams(session_row,MRI_MISSING_SCANS).split("|")
                            session_problem = getParams(session_row,MRI_PROBLEM_SCANS).split("|")

                            struct_poss  = cando_proc(CANDO_STRUCT_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_STRUCT,struct_poss)

                            dwi_poss = cando_proc(CANDO_DWI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_DWI,dwi_poss)

                            dwi_sdc_poss = cando_proc(CANDO_DWISDC_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_DWI_SDC,dwi_sdc_poss)

                            rsfmri_poss = cando_proc(CANDO_RSFMRI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_RSFMRI,rsfmri_poss)

                            rsfmri_sdc_poss = cando_proc(CANDO_RSFMRISDC_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_RSFMRI_SDC,rsfmri_sdc_poss)

                            rsfmri_dwi_sdc_poss = cando_proc(CANDO_RSFMRISDC_DWI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_RSFMRI_SDC_DWI,rsfmri_dwi_sdc_poss)

                            pasl_poss = cando_proc(CANDO_PASL_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASL,pasl_poss)

                            pasllegacy_poss = cando_proc(CANDO_PASLLEGACY_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASLLEGACY,pasllegacy_poss)

                            pcasl_poss = cando_proc(CANDO_PCASL_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PCASL,pcasl_poss)

                            pasl_sdc_poss = cando_proc(CANDO_PASLSDC_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASL_SDC,pasl_sdc_poss)

                            pasllegacy_sdc_poss = cando_proc(CANDO_PASLLEGACYSDC_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASLLEGACY_SDC,pasllegacy_sdc_poss)

                            pcasl_sdc_poss = cando_proc(CANDO_PCASLSDC_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PCASL_SDC,pcasl_sdc_poss)

                            pasl_fmri_sdc_poss = cando_proc(CANDO_PASLSDC_FMRI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASL_SDC_FMRI,pasl_fmri_sdc_poss)

                            pasllegacy_fmri_sdc_poss = cando_proc(CANDO_PASLLEGACYSDC_FMRI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASLLEGACY_SDC_FMRI,pasllegacy_fmri_sdc_poss)

                            pcasl_fmri_sdc_poss = cando_proc(CANDO_PCASLSDC_FMRI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PCASL_SDC_FMRI,pcasl_fmri_sdc_poss)

                            pasl_dwi_sdc_poss = cando_proc(CANDO_PASLSDC_DWI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASL_SDC_DWI,pasl_dwi_sdc_poss)

                            pasllegacy_dwi_sdc_poss = cando_proc(CANDO_PASLLEGACYSDC_DWI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PASLLEGACY_SDC_DWI,pasllegacy_dwi_sdc_poss)

                            pcasl_dwi_sdc_poss = cando_proc(CANDO_PCASLSDC_DWI_REQS,session_sequence,session_missing,problemscans=session_problem,struct_sessions=struct_sessions)
                            loadParams(session_row,CANDO_PCASL_SDC_DWI,pcasl_dwi_sdc_poss)

                        loadParams(session_row,CREATION_DATE,get_datetimestring_utc())

                    if not subject_t in excluded_participants:
                        table_data = appendTableRows(subject_row, table_data)


            bidsqc_df = pd.DataFrame(table_data, columns=table_header)
            bidsqc_df = bidsqc_df.sort_values(by=[HMLID])
            bidsqc_df.reset_index(drop=True, inplace=True)
            bidsqc_df.to_csv(csvout,index=False) 

            #bidsqc_mr_df = bidsqc_df[bidsqc_df["SESSION_MODALITY"] == "MR"]
            #bidsqc_mr_df.reset_index(drop=True, inplace=True)
            #bidsqc_mr_df.to_csv("./bidsqc/paninfo_release_250_bidsqc_mr",index=False) 

            # csv = "/xdisk/ryant/chidiugonna/Development/repos/PAN_Release_250/scripts/bidsqc/paninfo_release_250_bidsqc.csv"
            # bidsqc_df = df = pd.read_table(csv,sep=",")
            #subject_core_df = bidsqc_df[[PAN_RELEASE,HMLID, HMLID_REC, XNATID_ORIG,MCID,P2ID,P2SITE_ORIG,ORIG_PROJECT,GENDER,YOB]]
            #subject_cando_df = bidsqc_mr_df[[HMLID,CANDO_STRUCT,CANDO_DWI,CANDO_DWI_SDC,CANDO_PASL,CANDO_PASL_SDC,CANDO_PASL_SDC_FMRI,CANDO_PCASL,CANDO_PCASL_SDC,CANDO_PCASL_SDC_FMRI,CANDO_RSFMRI,CANDO_RSFMRI_SDC]]

            #subject_core_df = subject_core_df.drop_duplicates()

            subject_proc_df = bidsqc_df.groupby([HMLID]).apply(lambda s: pd.Series( { 
                "SUBJECT_" + CANDO_STRUCT : canProc(s[CANDO_STRUCT]),
                "SUBJECT_" +CANDO_DWI : canProc(s[CANDO_DWI]),
                "SUBJECT_" +CANDO_DWI_SDC : canProc(s[CANDO_DWI_SDC]),
                "SUBJECT_" +CANDO_PASL : canProc(s[CANDO_PASL]),
                "SUBJECT_" +CANDO_PASL_SDC : canProc(s[CANDO_PASL_SDC]),
                "SUBJECT_" +CANDO_PASL_SDC_FMRI : canProc(s[CANDO_PASL_SDC_FMRI]),
                "SUBJECT_" +CANDO_PASL_SDC_DWI : canProc(s[CANDO_PASL_SDC_DWI]),
                "SUBJECT_" +CANDO_PASLLEGACY : canProc(s[CANDO_PASLLEGACY]),
                "SUBJECT_" +CANDO_PASLLEGACY_SDC : canProc(s[CANDO_PASLLEGACY_SDC]),
                "SUBJECT_" +CANDO_PASLLEGACY_SDC_FMRI : canProc(s[CANDO_PASLLEGACY_SDC_FMRI]),
                "SUBJECT_" +CANDO_PASLLEGACY_SDC_DWI : canProc(s[CANDO_PASLLEGACY_SDC_DWI]),
                "SUBJECT_" +CANDO_PCASL : canProc(s[CANDO_PCASL]),
                "SUBJECT_" +CANDO_PCASL_SDC : canProc(s[CANDO_PCASL_SDC]),
                "SUBJECT_" +CANDO_PCASL_SDC_FMRI : canProc(s[CANDO_PCASL_SDC_FMRI]),
                "SUBJECT_" +CANDO_PCASL_SDC_DWI : canProc(s[CANDO_PCASL_SDC_DWI]),
                "SUBJECT_" +CANDO_RSFMRI : canProc(s[CANDO_RSFMRI]),
                "SUBJECT_" +CANDO_RSFMRI_SDC : canProc(s[CANDO_RSFMRI_SDC]),
                "SUBJECT_" +CANDO_RSFMRI_SDC_DWI : canProc(s[CANDO_RSFMRI_SDC_DWI]),
                "SUBJECT_" +CANDO_ASL : canProcList([s[CANDO_PASL],s[CANDO_PCASL],s[CANDO_PASLLEGACY]]),
                "SUBJECT_" +CANDO_ASL_SDC : canProcList([s[CANDO_PASL_SDC],s[CANDO_PCASL_SDC],s[CANDO_PASLLEGACY_SDC]]),
                "SUBJECT_" +CANDO_ASL_SDC_FMRI : canProcList([s[CANDO_PASL_SDC_FMRI],s[CANDO_PCASL_SDC_FMRI],s[CANDO_PASLLEGACY_SDC_FMRI]]),
                "SUBJECT_" +CANDO_ASL_SDC_DWI : canProcList([s[CANDO_PASL_SDC_DWI],s[CANDO_PCASL_SDC_DWI],s[CANDO_PASLLEGACY_SDC_DWI]])
            }))


            consolidated_df = pd.merge(bidsqc_df, subject_proc_df,  how='left', left_on=[HMLID], right_on = [HMLID])

            subject_csvout = os.path.splitext(csvout)[0] + "_subjectproc.csv"
            consolidated_df.to_csv(subject_csvout,index=False) 

            for header in consolidated_df.columns:
                if header.upper().startswith("MRIQC.") or header.upper().startswith("EDDYQC.") or header.upper().startswith("ASLPREPQC.") or header.upper().startswith("QSIPREPQC.") or '_AVAIL' in header.upper():
                    consolidated_df.pop(header) 

            subject_reduced_csvout = os.path.splitext(csvout)[0] + "_reduced_subjectproc.csv"
            consolidated_df.to_csv(subject_reduced_csvout,index=False)

            bidsqc_table_name = "xnat_mri_sessions.csv"
            if labels_dict and "BIDSQC_TABLENAME" in labels_dict.keys():
                bidsqc_table_name = getParams(labels_dict,"BIDSQC_TABLENAME")


            pan_mri_info_csvout = os.path.join(os.path.dirname(csvout),bidsqc_table_name)
            search_df = consolidated_df[(consolidated_df[SESSION_MODALITY]=="MR")]

            for header in search_df.columns:
                if 'CANDO' in header.upper() or 'RADNORMAL' in header.upper() or "MODALITY" in header.upper():
                    search_df.pop(header) 

            cols=list(search_df.columns.values)
            cols.pop(cols.index(HMLID))
            cols.pop(cols.index(BIDS_SUBJECT))
            cols.pop(cols.index(BIDS_SESSION))
            newdf=search_df[[HMLID] + [BIDS_SUBJECT] + [BIDS_SESSION] + cols]
            sorted_df = newdf.sort_values(by = [HMLID], ascending = [True])
            sorted_df.reset_index(drop=True,inplace=True)
            sorted_df.to_csv(pan_mri_info_csvout,sep=",", index=False)
            pan_mri_info_csvout_metadata = create_metadata(pan_mri_info_csvout,None, metadata = {"Script":"paninfo_bidsqc.py","Description":"Information about missing scans and scan issues"})

            if labels_dict:
                labels_dict["METADATA_FILE"]=pan_mri_info_csvout_metadata 
                labels_dict["OUTPUT_FILE"]=pan_mri_info_csvout
                export_labels(labels_dict,pipeline_config_file)


    except Exception as e:
        message = f"problem identifying BIDSQC files in {project.id} :  %s.\n{e}" 
        print (message)
        ACTION_OK=False

def main():
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()
    csvout = str(args.csvout)

    host = cleanServer(args.host)

    if args.projects:
        projects = args.projects
    else:
        projects = []

    if args.excluded_participants:
        excluded_participants = args.excluded_participants
    else:
        excluded_participants = []

    pipeline_config_file = None
    if args.pipeline_config_file:
        if Path(args.pipeline_config_file).exists():
            pipeline_config_file = str(args.pipeline_config_file)


    credentials = args.credentials
    user=None
    password=None
    if credentials:
        with open(credentials, 'r') as infile:
            cred_dict = json.load(infile)
            user = getParams(cred_dict,"user")
            password = getParams(cred_dict,"password")

    if user and password:
        print(f"user defined as {user} in credentials file {credentials}. Password also provided in same file.")
    elif user and not password:
        print(f"user defined as {user} in credentials file {credentials}. Password not provided in file. please enter password to match user.")
        password = getpass.getpass()
    elif not user and password:
        print(f"user not defined ain credentials file {credentials}. However a password was supplied. Please provide user that matches password.")
        user = input("User: ")
    elif not user and not password:
        print(f"Please provide user and password credentials.")
        user = input("User: ")
        password = getpass.getpass()

    getBidsQC(host,user,password,projects,csvout,excluded_participants=excluded_participants,pipeline_config_file=pipeline_config_file)


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()
